import gzip
import json
import math
import re
from enum import Enum
from pathlib import Path
import statistics
import logging

import datetime
from datetime import timezone
import requests
from pathlib import Path

from openpecha.core.annotation import Page, Span
from openpecha.core.annotations import BaseAnnotation, Language, OCRConfidence
from openpecha.core.layer import Layer, LayerEnum, OCRConfidenceLayer
from openpecha.core.ids import get_base_id
from openpecha.core.metadata import InitialPechaMetadata, InitialCreationType, LicenseType, Copyright_copyrighted, Copyright_unknown, Copyright_public_domain
from openpecha.formatters import BaseFormatter
from openpecha.utils import dump_yaml, gzip_str

from openpecha.buda.api import get_buda_scan_info, get_image_list, image_group_to_folder_name

ANNOTATION_MINIMAL_LEN = 20
ANNOTATION_MINIMAL_CONFIDENCE = 0.8
ANNOTATION_MAX_LOW_CONF_PER_PAGE = 10
NOISE_PATTERN = re.compile(r'(?:\s|[-，.… }#*,+•：/|(©:;་"།=@%༔){"])+')
GOOGLE_OCR_IMPORT_VERSION = "1.0.0"


class BBox:
    def __init__(self, text: str, vertices: list, confidence: float, language: str):

        self.text = text
        self.vertices = vertices
        self.confidence = confidence
        self.language = language
        self.mid_y = self.get_mid()
    
    def get_height(self):
        y1 = self.vertices[0][1]
        y2 = self.vertices[2][1]
        return y2 - y1
    
    def get_box_orientation(self):
        x1= self.vertices[0][0]
        x2 = self.vertices[1][0]
        y1= self.vertices[0][1]
        y2 = self.vertices[1][1]
        width = abs(x2-x1)
        length = abs(y2-y1)
        if width > length:
            return "landscape"
        else:
            return "portrait"

    def get_mid(self):
        """Calculate middle of the bounding poly vertically using y coordinates of the bounding poly

        Args:
            bounding_poly (dict): bounding poly's details

        Returns:
            float: mid point's y coordinate of bounding poly
        """
        y1 = self.vertices[0][1]
        y2 = self.vertices[2][1]
        y_avg = (y1 + y2) / 2
        return y_avg

    def get_centriod(self):
        """Calculate centriod of bounding poly

        Args:
            bounding_poly (dict): info regarding bounding poly such as vertices and description

        Returns:
            list: centriod coordinates
        """
        sum_of_x = 0
        sum_of_y = 0
        for vertice in self.vertices:
            sum_of_x += vertice[0]
            sum_of_y += vertice[1]
        return [sum_of_x/4, sum_of_y/4]


class GoogleVisionBDRCFileProvider():
    def __init__(self, bdrc_scan_id, ocr_import_info, ocr_disk_path=None, mode="local"):
        # ocr_base_path should be the output/ folder in the case of BDRC OCR files
        self.ocr_import_info = ocr_import_info
        self.ocr_disk_path = ocr_disk_path
        self.bdrc_scan_id = bdrc_scan_id
        self.mode = mode

    def get_image_list(self, image_group_id):
        buda_il = get_image_list(self.bdrc_scan_id, image_group_id)
        # format should be a list of image_id (/ file names)
        return map(lambda ii: ii["filename"], buda_il)

    def get_source_info(self):
        return get_buda_scan_info(self.bdrc_scan_id)

    def get_image_data(self, image_group_id, image_id):
        # TODO: implement the following modes:
        #  - "s3" (just read images from s3)
        #  - "s3-localcache" (cache s3 files on disk)
        # TODO: handle case where only zip archives are present on s3, one per volume.
        #       This should be indicated in self.ocr_import_info["ocr_info"]
        vol_folder = image_group_to_folder_name(self.bdrc_scan_id, image_group_id)
        expected_ocr_filename = image_id[:image_id.rfind('.')]+".json.gz"
        image_ocr_path = ocr_disk_path / vol_folder / expected_ocr_filename
        ocr_object = None
        try:
            ocr_object = json.load(gzip.open(str(expected_ocr_path), "rb"))
        except:
            logging.exception("could not read "+str(expected_ocr_path))
        return ocr_object

class GoogleVisionFormatter(BaseFormatter):
    """
    OpenPecha Formatter for Google OCR JSON output of scanned pecha.
    """

    def __init__(self, output_path=None, metadata=None):
        super().__init__(output_path, metadata)
        self.n_page_breaker_char = 3
        self.page_break = "\n" * self.n_page_breaker_char
        self.base_text = []
        self.low_conf_ann_text = ""
        self.base_meta = {}
        self.cur_base_word_confidences = []
        self.word_confidences = []
        self.bdrc_scan_id = None
        self.metadata = {}
        self.default_language = None
        self.source_info = {}
        self.remove_non_character_lines = True
        self.create_language_layer = True
        self.ocr_confidence_threshold = ANNOTATION_MINIMAL_CONFIDENCE
        self.language_annotation_min_len = ANNOTATION_MINIMAL_LEN

    def text_preprocess(self, text):
        return text


    def get_avg_bounding_poly_height(self, bounding_polys):
        """Calculate the average height of bounding polys in page

        Args:
            bounding_polys (list): list of boundingpolys

        Returns:
            float: average height of bounding ploys
        """
        height_sum = 0
        for bounding_poly in bounding_polys:
            height_sum += bounding_poly.get_height()
        avg_height = height_sum / len(bounding_polys)
        return avg_height

    def is_in_cur_line(self, prev_bounding_poly, bounding_poly, avg_height):
        """Check if bounding poly is in same line as previous bounding poly
        a threshold to check the conditions set to 10 but it can varies for pecha to pecha

        Args:
            prev_bounding_poly (dict): previous bounding poly
            bounding_poly (dict): current bounding poly
            avg_height (float): average height of all the bounding polys in page

        Returns:
            boolean: true if bouding poly is in same line as previous bounding poly else false
        """
        threshold = 10
        if (
            (bounding_poly.mid_y - prev_bounding_poly.mid_y)
            < (avg_height / threshold)
        ):
            return True
        else:
            return False

    def get_poly_lines(self, bounding_polys):
        """Return list of lines in page using bounding polys of page

        Args:
            bounding_polys (list): list of all the bounding polys

        Returns:
            list: list of lines in page
        """
        lines = []
        cur_line_polys = []
        prev_bounding_poly = bounding_polys[0]
        avg_line_height = self.get_avg_bounding_poly_height(bounding_polys)
        for bounding_poly in bounding_polys:
            if self.is_in_cur_line(prev_bounding_poly, bounding_poly, avg_line_height):
                cur_line_polys.append(bounding_poly)
            else:
                lines.append(cur_line_polys)
                cur_line_polys  = []
                cur_line_polys.append(bounding_poly)
            prev_bounding_poly = bounding_poly
        if cur_line_polys:
            lines.append(cur_line_polys)
        return lines


    def get_poly_sorted_on_y(self, bounding_poly_centriods):
        """Sort bounding polys centriod base on y coordinates

        Args:
            bounding_poly_centriods (list): list of centriod coordinates

        Returns:
            list: list centriod coordinated sorted base on y coordinate
        """
        sorted_on_y_polys = sorted(bounding_poly_centriods , key=lambda k: [k[1]])
        return sorted_on_y_polys

    def get_poly_sorted_on_x(self, sorted_on_y_polys, avg_box_height):
        """Groups box belonging in same line using average height and sort the grouped boxes base on x coordinates

        Args:
            sorted_on_y_polys (list): list of centriod
            avg_box_height (float): average boxes height

        Returns:
            list: list of sorted centriod
        """
        prev_bounding_poly = sorted_on_y_polys[0]
        lines = []
        cur_line = []
        sorted_polys = []
        for bounding_poly in sorted_on_y_polys:
            if abs(bounding_poly[1]-prev_bounding_poly[1]) < avg_box_height/10:
                cur_line.append(bounding_poly)
            else:
                lines.append(cur_line)
                cur_line = []
                cur_line.append(bounding_poly)
            prev_bounding_poly = bounding_poly
        if cur_line:
            lines.append(cur_line)
        for line in lines:
            sorted_line = sorted(line, key=lambda k: [k[0]])
            for poly in sorted_line:
                sorted_polys.append(poly)
        return sorted_polys

    def sort_bounding_polys(self, main_region_bounding_polys):
        """Sort the bounding polys

        Args:
            main_region_bounding_polys (list): list of bounding polys

        Returns:
            list: sorted bounding poly list
        """
        bounding_polys = {}
        bounding_poly_centriods = []
        avg_box_height = self.get_avg_bounding_poly_height(main_region_bounding_polys)
        for bounding_poly in main_region_bounding_polys:
            centroid = bounding_poly.get_centriod()
            # TODO: I'm not so sure about that...
            bounding_polys[f"{centroid[0]},{centroid[1]}"] = bounding_poly
            bounding_poly_centriods.append(centroid)
        sorted_bounding_polys = []
        sort_on_y_polys = self.get_poly_sorted_on_y(bounding_poly_centriods)
        sorted_bounding_poly_centriods = self.get_poly_sorted_on_x(sort_on_y_polys, avg_box_height)
        for bounding_poly_centriod in sorted_bounding_poly_centriods:
            sorted_bounding_polys.append(bounding_polys[f"{bounding_poly_centriod[0]},{bounding_poly_centriod[1]}"])
        return sorted_bounding_polys

    def has_space_attached(self, symbol):
        """Checks if symbol has space followed by it or not

        Args:
            symbol (dict): symbol info

        Returns:
            boolean: True if the symbol has space followed by it
        """
        if ('property' in symbol and 
                'detectedBreak' in symbol['property'] and 
                'type' in symbol['property']['detectedBreak'] and 
                symbol['property']['detectedBreak']['type'] == "SPACE"):
            return True
        return False
    
    def extract_vertices(self, word):
        """Extract vertices of bounding poly

        Args:
            word (dict): bounding poly info of a word

        Returns:
            list: list of vertices coordinates
        """
        vertices = []
        for vertice in word['boundingBox']['vertices']:
            vertices.append([vertice.get('x', 0), vertice.get('y', 0)])
        return vertices

    def dict_to_bbox(self, word):
        """Convert bounding poly to BBox object

        Args:
            word (dict): bounding poly of a word infos

        Returns:
            obj: BBox object of bounding poly
        """
        text = word.get('text', '')
        confidence = word.get('confidence')
        language = self.get_language_code(word)
        vertices = self.extract_vertices(word)
        bbox = BBox(text=text, vertices=vertices, confidence=confidence, language=language)
        return bbox

    def get_char_base_bounding_polys(self, response):
        """Return bounding polys in page response

        Args:
            response (dict): google ocr output of a page

        Returns:
            list: list of BBox object which saves required info of a bounding poly
        """
        bounding_polys = []
        cur_word = ""
        for page in response['fullTextAnnotation']['pages']:
            for block in page['blocks']:
                for paragraph in block['paragraphs']:
                    for word in paragraph['words']:
                        for symbol in word['symbols']:
                            cur_word += symbol['text']
                            if self.has_space_attached(symbol):
                                cur_word += " "
                        word['text'] = cur_word
                        cur_word = ""
                        bbox = self.dict_to_bbox(word)
                        bounding_polys.append(bbox)
        return bounding_polys

    def populate_confidence(self, bounding_polys):
        """Populate confidence of bounding polys of pecha level and image group level

        Args:
            bounding_polys (list): list of bounding polys
        """
        for bounding_poly in bounding_polys:
            self.word_confidences.append(float(bounding_poly.confidence))
            self.cur_base_word_confidences.append(float(bounding_poly.confidence))
    
    def get_space_poly_vertices(self, cur_bounding_poly, next_poly):
        """Return space bounding poly's vertices of located in between given polys

        Args:
            cur_bounding_poly (bbox): first bbox obj
            next_poly (bbox): second bbox obj

        Returns:
            list: list of vertices of space poly located in between
        """
        vertices = [
            cur_bounding_poly.vertices[1],
            next_poly.vertices[0],
            next_poly.vertices[3],
            cur_bounding_poly.vertices[2],
        ]
        return vertices
        
    def has_space_after(self, cur_bounding_poly, next_poly, avg_char_width):
        """Checks if there is space between two poly or not if yes returns a space poly bbox object

        Args:
            cur_bounding_poly (bbox): first bbox
            next_poly (bbox): second bbox
            avg_char_width (float): avg width of char in page

        Returns:
            bbox or none: space bbox object
        """
        cur_poly_top_right_corner = cur_bounding_poly.vertices[1][0]
        next_poly_top_left_corner = next_poly.vertices[0][0]
        if next_poly_top_left_corner - cur_poly_top_right_corner > avg_char_width*0.75 and cur_bounding_poly.text[-1] not in [' ', '་']:
            space_poly_vertices = self.get_space_poly_vertices(cur_bounding_poly, next_poly)
            space_box = BBox(
                text=" ",
                vertices=space_poly_vertices,
                confidence=None,
                language=None
            )
            return space_box
        return None
    
    def insert_space_bounding_poly(self, bounding_polys, avg_char_width):
        """Inserts space bounding poly if missing

        Args:
            bounding_polys (list): list of bbox objects
            avg_char_width (float): avg width of char

        Returns:
            list: list of bbox objects
        """
        new_bounding_polys = []
        for poly_walker, cur_bounding_poly in enumerate(bounding_polys):
            if poly_walker == len(bounding_polys)-1:
                new_bounding_polys.append(cur_bounding_poly)
            else:
                next_poly = bounding_polys[poly_walker+1]
                space_poly = self.has_space_after(cur_bounding_poly, next_poly, avg_char_width)
                if space_poly:
                    new_bounding_polys.append(cur_bounding_poly)
                    new_bounding_polys.append(space_poly)
                else:
                    new_bounding_polys.append(cur_bounding_poly)
        return new_bounding_polys
    
    def is_tibetan_non_consonant(self, symbol):
        """Checks if tibetan character is Tibetan but not a consonant

        Args:
            symbol (dict): symbol info

        Returns:
            boolean: True if character is Tibetan non-consonant
        """
        # we assume there is just one character:
        c = ord(symbol['text'][0])
        if (c >= ord('ༀ') and c <= ord('༿')) or (c >= ord('ཱ') and c <= ord('࿚')):
            return True

    def get_avg_char_width(self, response):
        """Calculate average width of box in a page, ignoring non consonant tibetan char

        Args:
            response (dict): ocr output of a page

        Returns:
            float: average box width in which character are saved
        """
        widths = []
        for page in response['fullTextAnnotation']['pages']:
            for block in page['blocks']:
                for paragraph in block['paragraphs']:
                    for word in paragraph['words']:
                        for symbol in word['symbols']:
                            # TODO: replace with proper diacritics detection
                            if self.is_tibetan_non_consonant(symbol):
                                continue
                            vertices = symbol['boundingBox']['vertices']
                            x1 = vertices[0].get('x', 0)
                            x2 = vertices[1].get('x', 0)
                            width = x2-x1
                            widths.append(width)
        return sum(widths) / len(widths)

    def save_boundingPoly(self, response, path):
        def tlbr(vertices):
            return {"tl": vertices[0], "br": vertices[2]}

        boundingPolies = {}
        for ann in response["textAnnotations"]:
            boundingPolies[ann["description"]] = tlbr(ann["boundingPoly"]["vertices"])

        zipped_boundingPolies = gzip_str(json.dumps(boundingPolies))
        path.write_bytes(zipped_boundingPolies)

    def save_low_conf_char(self, response, path):
        char_idx = 0
        low_conf_chars = ""
        for page in response["fullTextAnnotation"]["pages"]:
            for block in page["blocks"]:
                for paragraph in block["paragraphs"]:
                    for word in paragraph["words"]:
                        for symbol in word["symbols"]:
                            if symbol.get("confidence", 1) < self.ocr_confidence_threshold:
                                low_conf_chars += f'{symbol["text"]} {char_idx}\n'
                            char_idx += 1
        if low_conf_chars:
            path.write_bytes(gzip_str(low_conf_chars))

    def get_language_code(self, poly):
        lang = ""
        properties = poly.get("property", {})
        if properties:
            languages = properties.get("detectedLanguages", [])
            if languages:
                lang = languages[0]['languageCode']
        if lang == "":
            # this is not always true but replacing it with None is worse
            # with our current data
            return self.default_language
        if lang in ["bo", "en", "zh"]:
            return lang
        if lang == "dz":
            return "bo"
        # English is a kind of default for our purpose
        return "en"

    def add_language(self, poly, poly_start_cc, state):
        poly_lang = poly.language
        previous_ann = state["latest_language_annotation"]
        poly_end_cc = state["base_layer_len"] # by construction
        if previous_ann is not None:
            # if poly has the same language as the latest annotation, we just lengthen the previous
            # annotation to include this poly:
            if poly_lang == previous_ann['lang'] or not poly_lang:
                previous_ann["end"] = poly_end_cc
                return
            # if poly is the default language, we just conclude the previous annotation
            if poly_lang == self.default_language:
                state["latest_language_annotation"] = None
                return
            # else, we create a new annotation
            annotation = {"start": poly_start_cc, "end": poly_end_cc, "lang": poly_lang}
            state["language_annotations"].append(annotation)
            state["latest_language_annotation"] = annotation
            return
        # if there's no previous annotation and language is the default language, return
        if poly_lang == self.default_language or not poly_lang:
            return
        # if there's no previous annotation and language is not the default, we create an annotation
        annotation = {"start": poly_start_cc, "end": poly_end_cc, "lang": poly_lang}
        state["language_annotations"].append(annotation)
        state["latest_language_annotation"] = annotation

    def add_low_confidence(self, poly, poly_start_cc, state):
        if poly.confidence is None:
            return
        if poly.confidence > self.ocr_confidence_threshold:
            state["latest_low_confidence_annotation"] = None
            return
        poly_end_cc = state["base_layer_len"] # by construction
        if state["latest_low_confidence_annotation"] is not None:
            # average of the confidence indexes, weighted by character length
            state["latest_low_confidence_annotation"]["weights"].append((poly_end_cc - poly_start_cc, poly.confidence))
            state["latest_low_confidence_annotation"]["end"] = poly_end_cc
        else:
            annotation = {"start": poly_start_cc, "end": poly_end_cc, 
                "weights": [(poly_end_cc - poly_start_cc, poly.confidence)]}
            state["page_low_confidence_annotations"].append(annotation)
            state["latest_low_confidence_annotation"] = annotation

    def poly_line_has_characters(self, poly_line):
        for poly in poly_line:
            if not NOISE_PATTERN.match(poly.text):
                return True
        return False

    def build_page(self, ocr_object, image_number, image_filename, state):
        try:
            page_content = ocr_object["textAnnotations"][0]["description"]
        except Exception:
            logging.error("OCR page is empty (no textAnnotations[0]/description)")
            return
        avg_char_width = self.get_avg_char_width(ocr_object)
        bounding_polys = self.get_char_base_bounding_polys(ocr_object)
        sorted_bounding_polys = self.sort_bounding_polys(bounding_polys)
        sorted_bounding_polys = self.insert_space_bounding_poly(sorted_bounding_polys, avg_char_width)
        poly_lines = self.get_poly_lines(sorted_bounding_polys)
        page_start_cc = state["base_layer_len"]
        page_word_confidences = []
        for poly_line in poly_lines:
            if self.remove_non_character_lines and not self.poly_line_has_characters(poly_line):
                continue
            for poly in poly_line:
                state["base_layer"] += poly.text
                start_cc = state["base_layer_len"]
                state["base_layer_len"] += len(poly.text)
                if poly.confidence is not None:
                    state["word_confidences"].append(float(poly.confidence))
                    page_word_confidences.append(float(poly.confidence))
                    self.add_low_confidence(poly, start_cc, state)
                self.add_language(poly, start_cc, state)
            # adding a line break at the end of a line
            state["base_layer"] += "\n"
            state["base_layer_len"] += 1
        # if the whole page is below the min confidence level, we just add one
        # annotation for the page instead of annotating each word
        if page_word_confidences:
            mean_page_confidence = statistics.mean(page_word_confidences)
        else:
            mean_page_confidence = 0
        nb_below_threshold = len(state["page_low_confidence_annotations"])
        if mean_page_confidence < self.ocr_confidence_threshold or nb_below_threshold > self.max_low_conf_per_page:
            state["low_confidence_annotations"][self.get_unique_id()] = OCRConfidence(
                span=Span(start=page_start_cc, end=state["base_layer_len"]), 
                confidence=mean_page_confidence, nb_below_threshold=nb_below_threshold)
        else:
            self.merge_page_low_confidence_annotations(state["page_low_confidence_annotations"], state["low_confidence_annotations"])
            state["page_low_confidence_annotations"] = []
        # add pagination annotation:
        state["pagination_annotations"][self.get_unique_id()] = Page(
            span=Span(start=page_start_cc, end=state["base_layer_len"]), 
            imgnum=image_number, 
            reference=image_filename)
        # adding another line break at the end of a page
        state["base_layer"] += "\n"
        state["base_layer_len"] += 1

    def confidence_index_from_weighted_list(self, weights):
        sum_weights = 0
        confidence_sum = 0
        for weight, confidence in weights:
            sum_weights += weight
            confidence_sum += weight * confidence
        return confidence_sum / sum_weights

    def merge_page_low_confidence_annotations(self, annotation_list_src, annotations_dst):
        for annotation in annotation_list_src:
            avg_confidence = self.confidence_index_from_weighted_list(annotation["weights"])
            annotation_obj = OCRConfidence(
                span = Span(start=annotation["start"], end=annotation["end"]),
                confidence = avg_confidence)
            annotations_dst[self.get_unique_id()] = annotation_obj

    def merge_short_language_annotations(self, annotation_list):
        annotations = {}
        previous_annotation = None
        # annotation list is in span order
        for annotation in annotation_list:
            if annotation['end'] - annotation['start'] < self.language_annotation_min_len:
                if previous_annotation is not None and annotation['start'] - previous_annotation.span.end < self.language_annotation_min_len:
                    previous_annotation.span.end = annotation['end']
                continue
            if previous_annotation is not None and annotation["lang"] == previous_annotation.language:
                if annotation['start'] - previous_annotation.span.end < self.language_annotation_min_len:
                    previous_annotation.span.end = annotation['end']
                    continue
            previous_annotation = Language(
                span = Span(start=annotation["start"], end=annotation["end"]),
                language = annotation["lang"])
            annotations[self.get_unique_id()] = previous_annotation
        return annotations

    def build_base(self, image_group_id):
        """ The main function that takes the OCR results for an entire volume
            and creates its base and layers
        """
        image_list = self.data_provider.get_image_list(image_group_id)
        state = {
            "base_layer_len": 0,
            "base_layer": "",
            "low_confidence_annotations": {},
            "language_annotations": [],
            "pagination_annotations": {},
            "word_confidences": [],
            "latest_language_annotation": None,
            "latest_low_confidence_annotation": None,
            "page_low_confidence_annotations": []
        }
        for image_number, image_filename in enumerate(image_list):
            # enumerate starts at 0 but image numbers start at 1
            ocr_object = self.data_provider.get_image_data(image_group_id, image_filename)
            self.build_page(ocr_object, image_number+1, image_filename, state)
        layers = {}
        if state["pagination_annotations"]:
            layer = Layer(annotation_type=LayerEnum.pagination, annotations=state["pagination_annotations"])
            layers[LayerEnum.pagination.value] = json.loads(layer.json(exclude_none=True))
        if state["language_annotations"]:
            annotations = self.merge_short_language_annotations(state["language_annotations"])
            layer = Layer(annotation_type=LayerEnum.language, annotations=annotations)
            layers[LayerEnum.language.value] = json.loads(layer.json(exclude_none=True))
        if state["low_confidence_annotations"]:
            layer = OCRConfidenceLayer(
                confidence_threshold=self.ocr_confidence_threshold, 
                annotations=state["low_confidence_annotations"]
            )
            layers[LayerEnum.ocr_confidence.value] = json.loads(layer.json(exclude_none=True))

        return state["base_layer"], layers, state["word_confidences"]

    def get_copyright_and_license_info(self, bdata):
        if "copyright_status" not in bdata["source_metadata"]:
            return {}, None
        cs = bdata["source_metadata"]["copyright_status"]
        if cs == "http://purl.bdrc.io/resource/CopyrightPublicDomain":
            return Copyright_public_domain, LicenseType.CC0
        if cs == "http://purl.bdrc.io/resource/CopyrightUndetermined":
            return Copyright_unknown, None
        return Copyright_copyrighted, LicenseType.UNDER_COPYRIGHT
            
    def get_metadata(self, pecha_id, ocr_import_info):
        source_metadata = {
            "id": f"http://purl.bdrc.io/resource/{self.bdrc_scan_id}",
            "title": "",
            "author": "",
        }
        copyright = {}
        license = None
        if self.source_info is not None:
            source_metadata = self.source_info["source_metadata"]
            copyright, license = self.get_copyright_and_license_info(self.source_info)

        parser_link = ocr_import_info["parser_link"] if "parser_link" in ocr_import_info else None

        metadata = InitialPechaMetadata(
            id=pecha_id,
            source='https://library.bdrc.io',
            initial_creation_type=InitialCreationType.ocr,
            imported=datetime.datetime.now(timezone.utc),
            last_modified=datetime.datetime.now(timezone.utc),
            parser=parser_link,
            copyright=copyright,
            license=license,
            source_metadata=source_metadata,
            default_language=self.default_language,
            ocr_import_info=ocr_import_info
        )
        return json.loads(metadata.json())

    def set_base_meta(self, image_group_id, base_file_name, word_confidence_list):
        self.cur_word_confidences = []
        self.base_meta[base_file_name] = {
            "source_metadata": self.source_info["image_groups"][image_group_id],
            "order": self.source_info["image_groups"][image_group_id]["volume_number"],
            "base_file": f"{base_file_name}.txt"
            }
        if word_confidence_list:
            self.base_meta[base_file_name]["statistics"] = {
              "ocr_word_median_confidence_index": statistics.median(word_confidence_list),
              "ocr_word_mean_confidence_index": statistics.mean(word_confidence_list)
            }
    
    def create_opf(self, data_provider, pecha_id, opf_options = {}, ocr_import_info = {}):
        """Create opf of google ocred pecha

        Args:
            data_provider (DataProvider): an instance that will be used to get the necessary data
            pecha_id (str): pecha id
            opf_options (Dict): an object with the following keys:
                create_language_layer: boolean
                language_annotation_min_len: int
                ocr_confidence_threshold: float (use -1.0 for no OCR confidence layer)
                remove_non_character_lines: boolean
                max_low_conf_per_page: int
            ocr_import_info (Dict): an object with the following keys:
                bdrc_scan_id: str
                source: str
                ocr_info: Dict
                batch_id: str
                software_id: str
                expected_default_language: str

        Returns:
            path: opf path
        """

        self.data_provider = data_provider

        self.remove_non_character_lines = opf_options["remove_non_character_lines"] if "remove_non_character_lines" in opf_options else True
        self.create_language_layer = opf_options["create_language_layer"] if "create_language_layer" in opf_options else True
        self.ocr_confidence_threshold = opf_options["ocr_confidence_threshold"] if "ocr_confidence_threshold" in opf_options else ANNOTATION_MINIMAL_CONFIDENCE
        self.language_annotation_min_len = opf_options["language_annotation_min_len"] if "language_annotation_min_len" in opf_options else ANNOTATION_MINIMAL_LEN
        self.max_low_conf_per_page = opf_options["max_low_conf_per_page"] if "max_low_conf_per_page" in opf_options else ANNOTATION_MAX_LOW_CONF_PER_PAGE

        ocr_import_info["op_import_options"] = opf_options
        ocr_import_info["op_import_version"] = GOOGLE_OCR_IMPORT_VERSION

        self._build_dirs(None, id_=pecha_id)

        # if the bdrc scan id is not specified, we assume it's the directory namepecha_id
        self.bdrc_scan_id = ocr_import_info["bdrc_scan_id"]
        self.source_info = data_provider.get_source_info()
        self.default_language = "bo" if "expected_default_language" not in ocr_import_info else ocr_import_info["expected_default_language"]
        self.default_language = "bo"
        if "expected_default_language" in ocr_import_info:
            self.default_language = ocr_import_info["expected_default_language"]
        elif "languages" in buda_data and buda_data["languages"]:
            self.default_language = buda_data["languages"][0]

        self.metadata = self.get_metadata(pecha_id, ocr_import_info)
        total_word_confidence_list = []

        for image_group_id, image_group_info in self.source_info["image_groups"].items():
            base_id = image_group_id
            base_text, layers, word_confidence_list = self.build_base(image_group_id)

            # save base_text
            (self.dirs["opf_path"] / "base" / f"{base_id}.txt").write_text(base_text)

            # save layers
            vol_layer_path = self.dirs["layers_path"] / base_id
            vol_layer_path.mkdir(exist_ok=True)
            for layer_id, layer in layers.items():
                layer_fn = vol_layer_path / f"{layer_id}.yml"
                dump_yaml(layer, layer_fn)
            self.set_base_meta(image_group_id, base_id, word_confidence_list)
            total_word_confidence_list += word_confidence_list

        # we add the rest to metadata:
        self.metadata["bases"] = self.base_meta
        if total_word_confidence_list:
            self.metadata['statistics'] = {
                # there are probably more efficient ways to compute those
                "ocr_word_mean_confidence_index": statistics.mean(total_word_confidence_list),
                "ocr_word_median_confidence_index": statistics.median(total_word_confidence_list)
            }

        meta_fn = self.dirs["opf_path"] / "meta.yml"
        dump_yaml(self.metadata, meta_fn)

        return self.dirs["opf_path"].parent



import re
from pathlib import Path

import yaml

from openpecha.formatters.layers import AnnType

from .serialize import Serialize


class PedurmaSerializer(Serialize):
    """Pedurma serializer class to get diplomatic text."""

    def __get_adapted_span(self, span, vol_id):
        """Adapts the annotation span to base-text of the text

        Adapts the annotation span, which is based on volume base-text
        to text base-text.

        Args:
            span (dict): span of a annotation, eg: {start:, end:}
            vol_id (str): id of vol, where part of the text exists.

        Returns:
            adapted_start (int): adapted start based on text base-text
            adapted_end (int): adapted end based on text base-text

        """
        adapted_start = span["start"] - self.text_spans[vol_id]["start"]
        adapted_end = span["end"] - self.text_spans[vol_id]["start"]
        return adapted_start, adapted_end

    def get_local_id(self, ann, uuid2localid):
        try:
            return chr(uuid2localid[ann["id"]])
        except Exception:
            return ""

    def apply_annotation(self, vol_id, ann, uuid2localid):
        """Applies annotation to specific volume base-text, where part of the text exists.

        Args:
            vol_id (str): id of vol, where part of the text exists.
            ann (dict): annotation of any type.

        Returns:
            None

        """
        only_start_ann = False
        start_payload = "("
        end_payload = ")"
        side = "ab"
        local_id = self.get_local_id(ann, uuid2localid)
        if ann["type"] == AnnType.pagination:
            pg_idx = ann.get("page_index", "")
            if not pg_idx:
                pg_idx = ann.get("imgnum", "")
            if pg_idx == "0b":
                pg_n = ann["reference"][5:-1]
                pg_side = ann["reference"][-1]
                if "-" in pg_n:
                    pg_n = int(pg_n.split("-")[0])
                    pg_side = side[int(pg_side)]
                    start_payload = f"[{local_id}{pg_n}{pg_side}]"
                else:
                    pg_n = int(pg_n)
                    if pg_side.isdigit():
                        pg_n = str(pg_n) + pg_side
                        pg_side = ""
                    start_payload = f"〔{local_id}{pg_n}{pg_side}〕"
            else:
                start_payload = f"〔{local_id}{pg_idx}〕"
            if ann.get("reference", ""):
                start_payload += f' {ann["reference"]}\n'
            else:
                start_payload += "\n"
            only_start_ann = True
        elif ann["type"] == AnnType.pedurma_note:
            start_payload = ""
            end_payload = f'{ann["collation_note"]}'

        start_cc, end_cc = self.__get_adapted_span(ann["span"], vol_id)
        self.add_chars(vol_id, start_cc, True, start_payload)
        if not only_start_ann:
            self.add_chars(vol_id, end_cc, False, end_payload)

    def get_chunks(self, text):
        """Break the text on where notes are located

        Args:
            text (text): preview text

        Returns:
            list: contain all the chunks and note along with it
        """
        result = []
        cur_note = []
        chunks = re.split(r"(\<.+?\>)", text)
        for chunk in chunks:
            if "{" in chunk:
                note = yaml.safe_load(chunk)
                cur_note.append(note)
                result.append(cur_note)
                cur_note = []
            else:
                cur_note.append(chunk)
        result.append([chunk, {}])
        return result

    def process_chunk(self, chunk, pub):
        """Reinsert the note to chunk according to publication selected

        Args:
            chunk (list): text chunk and its notes
            pub (str): publication annotation

        Returns:
            str: chunk text with note added according to publication
        """
        chunk_text = chunk[0]
        chunk_notes = chunk[1]
        if chunk_notes:
            note = chunk_notes[pub]
            if note:
                old_note = re.search(r"(:\S+)\n?", chunk_text).group(1)
                chunk_text = re.sub(old_note, note, chunk_text)
            else:
                chunk_text = re.sub(":", "", chunk_text)
        return chunk_text

    def get_diplomatic_text(self, text, pub):
        """Reinsert the notes from pedurma of given publication and generates version of the text of given publication

        Args:
            text (str): text with pedurma notes of all other publication
            pub (str): annotation of publication de for derge, pe for pecing, nar for narthang and co for cone

        Returns:
            str: Text of mentioned publication extracted from pedurma text
        """
        diplomatic_text = ""
        chunks = self.get_chunks(text)
        for chunk in chunks:
            diplomatic_text += self.process_chunk(chunk, pub)
        return diplomatic_text

    def serialize(self, output_path="./output/diplomatic_text/", pub="pe"):
        """Serialize pedurma preview opf to diplomatic text of given publication

        Args:
            output_path (str, optional): output path where user wants to save the diplomatic text. Defaults to "./output/diplomatic_text/".
            pub (str, optional): publication annotation . Defaults to 'pe'.
        """
        output_path = Path(output_path)
        self.apply_layers()
        results = self.get_result()
        for vol_id, result in results.items():
            # diplomatic_text = self.get_diplomatic_text(result, pub)
            (output_path / vol_id).write_text(result, encoding="utf-8")
        print("Serialize complete...")



import logging
import random


def _sample_text_chunk(text):
    start_idx = random.randint(0, 100000)
    text_chunk = text[start_idx : start_idx + 1000]
    first_syl_idx = text_chunk.find("་")
    last_syl_idx = text_chunk.rfind("་")
    return text_chunk[first_syl_idx + 1 : last_syl_idx]


def is_text_good_quality(text, strategy=None, threshold=0.2):
    if not strategy:
        try:
            from bonltk.text_quality import non_words_ratio

            strategy = non_words_ratio
        except Exception:
            msg = (
                "bonltk not installed. Install it with `pip install bonltk` "
                "or from https://github.com/10zinten/bonltk"
            )
            raise ImportError(msg)
    text_chunk = _sample_text_chunk(text)
    ratio = strategy(text_chunk)
    text_quality = 1 - ratio
    logging.info(f"Text quality: {text_quality}")
    if text_quality <= threshold:
        return True
    else:
        return False



from pathlib import Path

from openpecha.alignment.exporter import Exporter


class BitextExporter(Exporter):
    def serialize_segment_pair(self, pair_id, segment_pair, segment_texts):
        """Generate bitext segment using alignment information

        Args:
            pair_id (uuid): segment pair id
            segment_pair (dict): key as pecha id and value as segment id of the pecha
            segment_texts (dict): key as pecha id and value as another dict whose key as segment pair id and value as segment text

        Returns:
            str: bitext segment text where segments were separated by newline
        """
        segment_pair_text = ""
        for pair_c, (pecha_id, segment_id) in enumerate(segment_pair.items(), 1):
            segment_text = segment_texts[pecha_id][pair_id]
            segment_text = segment_text.replace("\n", "")
            if pair_c == 1:
                segment_pair_text += f"{segment_text}\n"
            else:
                segment_pair_text += f"\t{segment_text}\n"
        return segment_pair_text

    def export(self, output_file_path):
        """Export alignment to bitext

        Args:
            output_file_path (str): output file path
        """
        bi_text = ""
        segment_srcs = self.alignment.get("segment_sources", {})
        segment_texts = {}
        for seg_src_id, seg_src in segment_srcs.items():
            segment_texts[seg_src_id] = self.get_segment_texts(seg_src_id)
        segment_pairs = self.alignment.get("segment_pairs", {})
        for pair_id, segment_pair in segment_pairs.items():
            bi_text += self.serialize_segment_pair(pair_id, segment_pair, segment_texts)
        Path(output_file_path).write_text(bi_text, encoding="utf-8")



import os
import re
from pathlib import Path
from uuid import uuid4

import polib
from git import Repo

from openpecha import config
from openpecha.core.layer import Layer, LayerEnum
from openpecha.core.pecha import OpenPechaFS
from openpecha.github_utils import commit
from openpecha.utils import dump_yaml

from ...alignment.tmx import TMXAlignment
from ...alignment.tmx.create_opf import (
    create_readme,
    get_segment_annotation,
    publish_pecha,
)


def setup_auth(repo, org, token):
    remote_url = repo.remote().url
    old_url = remote_url.split("//")
    authed_remote_url = f"{old_url[0]}//{org}:{token}@{old_url[1]}"
    repo.remote().set_url(authed_remote_url)


def get_segment_layer_base_and_mapping(po_dict):
    segment_annotations = {}
    char_walker = 0
    base_text = ""
    annotation_map = {}
    curr_dict = {}
    for uid, text_info in po_dict.items():
        text = text_info["text"]
        segment_annotation, end = get_segment_annotation(text, char_walker)
        segment_annotations.update(segment_annotation)
        char_walker = end + 2
        annotation_id = list(segment_annotation.keys())[0]
        base_text += text + "\n"
        curr_dict[uid] = {"annotation_id": annotation_id}
        annotation_map.update(curr_dict)
        curr_dict = {}

    segment_layer = Layer(
        annotation_type=LayerEnum.segment, annotations=segment_annotations
    )
    return segment_layer, base_text, annotation_map


def create_opf_for_translated(po_dict, pecha_id, opf_path, title=None, lang=None):
    opf = OpenPechaFS(path=opf_path)
    segment_layer, base_text, annotation_map = get_segment_layer_base_and_mapping(
        po_dict
    )
    layers = {"0001": {LayerEnum.segment: segment_layer}}
    bases = {"0001": base_text}
    metadata = {
        "id": pecha_id,
        "initial_creation_type": "TMX",
        "origin_type": "translation",
        "source_metadata": {
            "id": "",
            "title": title,
            "langauge": lang,
            "author": "",
            "volume": {},
        },
    }
    opf.layers = layers
    opf.base = bases
    opf.save_base()
    opf.save_layers()
    meta_fn = Path(f"{opf_path}/meta.yml")
    dump_yaml(metadata, meta_fn)
    return annotation_map


def create_opf_and_annotation_map(po_zip, title=None, lang=None):
    if po_zip:
        pecha_id = uuid4().hex
        opf_path = Path(config.PECHAS_PATH / pecha_id / f"{pecha_id}.opf/")
        annotation_map = create_opf_for_translated(
            po_zip, pecha_id, opf_path, title, lang
        )
        readme = create_readme(opf_path.parent)
        Path(opf_path.parent / "readme.md").write_text(readme, encoding="utf-8")
    return opf_path.parent, annotation_map


def get_id_and_string(po_file):
    po_dict = {}
    curr_po = {}
    lang = po_file.metadata["Language"]
    for entry in po_file:
        if lang == "bo":
            id = entry.msgid
            string = entry.tcomment
        elif lang == "en":
            id = entry.msgid
            string = entry.msgstr
        text = re.sub(r"\n", " ", string)
        curr_po[id] = {"text": text}
        po_dict.update(curr_po)
        curr_po = {}
    return po_dict, lang


def read_po(path):
    return polib.pofile(path)


def transifex_output_to_opf(po_path):
    po_file = read_po(po_path)
    po_zip, lang = get_id_and_string(po_file)
    title = "this is the title"
    target_pecha_path, annotation_map = create_opf_and_annotation_map(
        po_zip, title, lang
    )
    return target_pecha_path, annotation_map


def update_alignment_from_po(po_path, alignment_path, publish=True):
    target_pecha_path, annotation_map = transifex_output_to_opf(po_path)
    obj = TMXAlignment()
    obj.update_alignment_repo(alignment_path, target_pecha_path, annotation_map)

    if publish:
        repo = Repo(alignment_path)
        token = os.environ.get("GITHUB_TOKEN")
        org = os.environ["OPENPECHA_DATA_GITHUB_ORG"]
        setup_auth(repo, org, token)
        message = "updated alginment.yml"
        commit(repo, message, not_includes=[], branch="master")
        publish_pecha(target_pecha_path)
    return target_pecha_path



from logging import NOTSET
from pathlib import Path

from openpecha.utils import download_pecha, load_yaml


class Exporter:
    def __init__(self, alignment_path) -> None:
        self.alignment_path = alignment_path if alignment_path else None
        self.alignment = load_yaml(self.alignment_path) if self.alignment_path else {}

    def get_base_layer(self, pecha_id, pecha_path=None):
        """Return base text of the pecha

        Args:
            pecha_id (st): pecha uuid
            pecha_path (str, optional): path of the pecha. Defaults to None.

        Returns:
            [str]: base text of the pecha
        """
        base = ""
        if pecha_path is None:
            pecha_path = download_pecha(pecha_id)
        base = (pecha_path / f"{pecha_id}.opf" / "base" / "0001.txt").read_text(
            encoding="utf-8"
        )
        return base

    def get_segment_layer(self, pecha_id, pecha_path):
        """Return src pecha's segment annotations detail from segment layer

        Args:
            pecha_id (str): pecha id
            pecha_path (str): pecha path

        Returns:
            dict: segment annotations detail
        """
        if pecha_path is None:
            pecha_path = download_pecha(pecha_id)
        try:
            segment_layer = load_yaml(
                (pecha_path / f"{pecha_id}.opf" / "layers" / "0001" / "Segment.yml")
            )
            segment_annotations = segment_layer["annotations"]
            return segment_annotations
        except Exception:
            return {}

    def get_segment(self, segment_ann, base_text):
        """Extract segment text and return

        Args:
            segment_ann (dict): segment annotation detail
            base_text (str): base text of the segment

        Returns:
            str: segment text from base text
        """
        segment_text = ""
        if segment_ann:
            segment_start = segment_ann["span"]["start"]
            segment_end = segment_ann["span"]["end"]
            segment_text = base_text[segment_start : segment_end + 1]
        return segment_text

    def get_segment_texts(self, pecha_id, pecha_path=None):
        segment_texts = {}
        if pecha_path is None:
            pecha_path = download_pecha(pecha_id)
        base_text = self.get_base_layer(pecha_id, pecha_path)
        segment_layer = self.get_segment_layer(pecha_id, pecha_path)
        for pair_id, segment_pair in self.alignment["segment_pairs"].items():
            segment_id = segment_pair.get(pecha_id, "")
            segment_ann = segment_layer.get(segment_id, {})
            if segment_ann:
                segment_texts[pair_id] = self.get_segment(segment_ann, base_text)
        return segment_texts

    def export(self):
        raise NotImplementedError



"""Module contains all the Annotations classes
"""

from typing import Dict, Optional

from pydantic import BaseModel, Extra, Field, validator

from .ids import get_uuid


class Span(BaseModel):
    start: int = Field(..., ge=0)
    end: int = Field(..., ge=0)
    errors: Optional[Dict] = None

    @validator("start", "end")
    def span_must_not_be_neg(cls, v):

        if v < 0:
            raise ValueError("span shouldn't be negative")
        return v

    @validator("end")
    def end_must_not_be_less_than_start(cls, v, values, **kwargs):
        if "start" in values and v < values["start"]:
            raise ValueError("Span end must not be less than start")
        return v

    class Config:
        extra = Extra.forbid


class BaseAnnotation(BaseModel):
    span: Span
    metadata: Optional[Dict] = None

    class Config:
        extra = Extra.forbid


class Pagination(BaseAnnotation):
    page_info: str = Field(None, description="page payload")
    imgnum: int = Field(None, description="image sequence number")
    order: int = Field(None, description="order of the page")
    reference: str = Field(
        None, description="can be url or just string indentifier of source page"
    )


class Language(BaseAnnotation):
    language: str = Field(None, description="BCP-47 tag of the language")

class OCRConfidence(BaseAnnotation):
    confidence: float
    nb_below_threshold: Optional[int]

class Citation(BaseAnnotation):
    pass


class Correction(BaseAnnotation):
    pass


class ErrorCandidate(BaseAnnotation):
    pass


class Pedurma(BaseAnnotation):
    pass


class Sabche(BaseAnnotation):
    pass


class Tsawa(BaseAnnotation):
    pass


class Yigchung(BaseAnnotation):
    pass


class Archaic(BaseAnnotation):
    pass


class Durchen(BaseAnnotation):
    default: str = Field(..., description="text_name of the default option")
    options: Dict[str, str] = Field(
        ..., description="all other spell options in dict of {text_name, option}"
    )
    printable: bool = True


class Footnote(BaseAnnotation):
    pass


class Segment(BaseAnnotation):
    pass



import re
from pathlib import Path

import polib
from botok.tokenizers.sentencetokenizer import get_normalized_sentence
from botok.tokenizers.wordtokenizer import WordTokenizer

from openpecha.alignment.exporter import Exporter


class PoExporter(Exporter):
    def __init__(self, alignment_path) -> None:
        super().__init__(alignment_path)
        self.file = polib.POFile()
        self.file.metadata = {
            "MIME-Version": "1.0",
            "Content-Type": "text/plain; charset=utf-8",
            "Content-Transfer-Encoding": "8bit",
        }

    def _create_entry(
        self, msgid, msgstr="", msgctxt=None, comment=None, tcomment=None
    ):
        """

        :param msgid: string, the entry msgid.
        :param msgstr: string, the entry msgstr.
        :param msgctxt: string, the entry context.
        :param comment: string, the entry comment.
        :param tcomment: string, the entry translator comment.
        """
        entry = polib.POEntry(
            msgid=msgid,
            msgstr=msgstr,
            msgctxt=msgctxt,
            comment=comment,
            tcomment=tcomment,
        )
        self.file.append(entry)

    def write_to_file(self, filename):
        """Save pofile object to mentioned filename

        Args:
            filename (path): file path
        """
        self.file.save(filename)

    def segment_to_entries(self, pecha_id, pecha_path, lang="bo"):
        """Add segment to po file

        Args:
            pecha_id (uuid): pecha id
            pecha_path (path): pecha path
            lang (str): language of the pecha
        """
        segments = self.get_segment_texts(pecha_id, pecha_path)
        wt = WordTokenizer()
        for pair_id, segment_text in segments.items():
            segment_text = segment_text.replace("\n", "")
            if lang == "bo":
                segment_tokens = wt.tokenize(segment_text, split_affixes=True)
                tokenized_segment = get_normalized_sentence(segment_tokens)
                self._create_entry(
                    msgid=pair_id, tcomment=segment_text, msgstr=tokenized_segment
                )
            else:
                self._create_entry(msgid=pair_id, msgstr=segment_text)

    def export(self, output_dir):
        """Export all the segment src po file in given directory

        Args:
            output_dir (str): output directory in which po files will be saved
        """
        segment_srcs = self.alignment.get("segment_sources", {})
        for seg_src_id, seg_src in segment_srcs.items():
            lang = seg_src.get("lang", "")
            self.file.metadata["Language"] = lang
            if lang:
                self.segment_to_entries(seg_src_id, pecha_path=None, lang=lang)
                self.write_to_file((Path(output_dir) / f"{lang}.po"))
            self.file = polib.POFile()
            self.file.metadata = {
                "MIME-Version": "1.0",
                "Content-Type": "text/plain; charset=utf-8",
                "Content-Transfer-Encoding": "8bit",
            }



import re
from pathlib import Path
import rdflib
from rdflib import BNode, Literal, URIRef
from rdflib.namespace import OWL, RDF, RDFS, SKOS, XSD, Namespace, NamespaceManager
from openpecha.buda.chunker import TibetanEasyChunker, EnglishEasyChunker
from openpecha.core.layer import Layer, LayerEnum, PechaMetadata, SpanINFO

rdf = RDF
rdfs = RDFS
bdr = Namespace("http://purl.bdrc.io/resource/")
bdo = Namespace("http://purl.bdrc.io/ontology/core/")
bdg = Namespace("http://purl.bdrc.io/graph/")
bda = Namespace("http://purl.bdrc.io/admindata/")
adm = Namespace("http://purl.bdrc.io/ontology/admin/")

nsm = NamespaceManager(rdflib.Graph())
nsm.bind("bdr", bdr)
nsm.bind("", bdo)
nsm.bind("bdg", bdg)
nsm.bind("bda", bda)
nsm.bind("adm", adm)
nsm.bind("rdf", rdf)
nsm.bind("rdfs", rdfs)

class BUDARDFSerializer:
    """
    """

    def __init__(self, openpecha):
        self.openpecha = openpecha
        self._pecha_id = openpecha.meta.id
        self.lname = f"IE0OP{self._pecha_id}"
        self.graph_r = bdg[self.lname]
        self.lod_ds = rdflib.Dataset()
        self.lod_g = self.lod_ds.graph(self.graph_r)
        self.lod_g.namespace_manager = nsm
        self.bl_volinfo = None

    def add_triple(self, rdf_subject, rdf_predicate, rdf_object):
        self.lod_g.add((rdf_subject, rdf_predicate, rdf_object))

    # implementation of the API of Serialize
    def apply_layers(self):
        self.set_instance()

    def get_result(self):
        return self.lod_g


    """
    Entry point to build the RDF graph
    """

    def set_instance(self):
        self.add_triple(bdr[f"{self.lname}"], rdf.type, bdo["EtextInstance"])
        meta = self.openpecha.meta
        sm = meta.source_metadata
        if sm is not None:
            scanlname = sm["id"]
            if scanlname.startswith("bdr:"):
                scanlname = scanlname[4:]
            elif scanlname.startswith("http://purl.bdrc.io/resource/"):
                scanlname = scanlname[29:]
            self.add_triple(
                bdr[self.lname], bdo["instanceReproductionOf"], bdr[scanlname]
            )
            self.add_triple(
                bdr[scanlname], bdo["instanceHasReproduction"], bdr[self.lname]
            )
        if meta.initial_creation_type == "ocr":
            self.add_triple(
                bdr[self.lname], bdo["contentMethod"], bdr['ContentMethod_OCR']
            )
        if meta.statistics is not None:
            if "ocr_word_median_confidence_index" in meta.statistics:
                self.add_triple(
                    bdr[self.lname], bdo["OPFOCRWordMedianConfidenceIndex"], Literal(meta.statistics["ocr_word_median_confidence_index"], datatype=XSD.float)
                )
            if "ocr_word_mean_confidence_index" in meta.statistics:
                self.add_triple(
                    bdr[self.lname], bdo["OPFOCRWordMeanConfidenceIndex"], Literal(meta.statistics["ocr_word_mean_confidence_index"], datatype=XSD.float)
                )
        if meta.legacy_id is not None and meta.legacy_id:
            legacylname = f"IE0OP{meta.legacy_id}"
            self.add_triple(bdr[legacylname], rdf.type, bdo["EtextInstance"])
            self.add_triple(bda[legacylname], rdf.type, adm["AdminData"])
            self.add_triple(bda[legacylname], adm["adminAbout"], bdr[legacylname])
            self.add_triple(bda[legacylname], adm["status"], bda['StatusWithdrawn'])
            self.add_triple(bda[legacylname], adm["replaceWith"], bdr[self.lname])
        if meta.ocr_import_info is not None:
            oii = meta.ocr_import_info
            if "source" in oii:
                self.add_triple(bdr[self.lname], bdo["OPFOCRSource"], Literal(oii["source"]))
            if "software" in oii:
                self.add_triple(bdr[self.lname], bdo["OPFOCRSoftware"], Literal(oii["software"]))
            if "batch" in oii:
                self.add_triple(bdr[self.lname], bdo["OPFOCRBatch"], Literal(oii["batch"]))
            if "ocr_info" in oii and "timestamp" in oii["ocr_info"]:
                self.add_triple(bdr[self.lname], bdo["OPFOCRTimeStamp"], Literal(oii["ocr_info"]["timestamp"], datatype=XSD.dateTime))
        self.get_base_volumes()
        self.set_adm()

    def get_base_volumes(self):
        for baselname, baseinfo in self.openpecha.meta.bases.items():
            volume_string = self.openpecha.get_base(baselname)
            if len(volume_string) < 2:
                continue
            volume_number = 0
            if "source_metadata" in baseinfo and "volume_number" in baseinfo["source_metadata"]:
                volume_number = baseinfo["source_metadata"]["volume_number"]
            elif "order" in baseinfo:
                volume_number = baseinfo["order"]
            else:
                volume_number = int(re.search(r"\d+", baselname).group())
            self.set_etext_asset(baselname, baseinfo, volume_number)
            self.add_triple(
                bdr[self.lname],
                bdo["instanceHasVolume"],
                bdr[f"VL{self.lname}_{baselname}"],
            )
            self.set_etext_ref(baselname)
            self.set_etext(baselname, baseinfo, volume_number)

    def set_etext_asset(self, baselname, baseinfo, volume_number):
        volume_basename = f"{self.lname}_{baselname}"
        subject = bdr[f"VL{volume_basename}"]
        self.add_triple(subject, rdf.type, bdo["VolumeEtextAsset"])
        self.add_triple(subject, bdo["volumeHasEtext"], bdr[f"ER{volume_basename}"])
        self.add_triple(
            subject, bdo["volumeNumber"], Literal(volume_number, datatype=XSD.integer)
        )
        self.add_triple(subject, bdo["volumeOf"], bdr[f"{self.lname}"])

    def set_etext_ref(self, baselname):
        volume_basename = f"{self.lname}_{baselname}"
        subject = bdr[f"ER{volume_basename}"]
        self.add_triple(subject, rdf.type, bdo["EtextRef"])
        self.add_triple(subject, bdo["eTextResource"], bdr[f"UT{volume_basename}"])
        self.add_triple(subject, bdo["seqNum"], Literal(1, datatype=XSD.integer))

    def set_etext(self, baselname, baseinfo, volume_number):
        volume_basename = f"{self.lname}_{baselname}"
        subject = bdr[f"UT{volume_basename}"]
        if "source_metadata" in baseinfo and "image_group_id" in baseinfo["source_metadata"]:
            iglname = baseinfo["source_metadata"]["image_group_id"]
            if iglname.startswith("bdr:"):
                iglname = iglname[4:]
            elif iglname.startswith("http://purl.bdrc.io/resource/"):
                iglname = iglname[29:]
            self.add_triple(subject, bdo["eTextForImageGroup"], bdr[iglname])
        self.add_triple(subject, rdf.type, bdo["Etext"])
        self.add_triple(subject, bdo["eTextInInstance"], bdr[self.lname])
        self.add_triple(
            subject, bdo["eTextIsVolume"], Literal(volume_number, datatype=XSD.integer)
        )
        self.add_triple(
            subject,
            rdfs.seeAlso,
            Literal(
                f"https://github.com/Openpecha-Data/{self._pecha_id}/", datatype=XSD.anyURI
            ),
        )
        self.set_etext_pages(baselname)
        self.set_etext_chunks(baselname, baseinfo)

    def set_etext_pages(self, baselname):
        player = self.openpecha.get_layer(baselname, LayerEnum.pagination)
        if player is None:
            return
        for annotation_id, annotation in player.annotations.items():
            self.set_etext_page(annotation_id, annotation, baselname)

    def set_etext_page(self, annotation_id, annotation, volume_name):
        volume_basename = f"{self.lname}_{volume_name}"
        subject = bdr[f'EP{annotation_id}']
        sequence = 0
        if "imgnum" in annotation:
            sequence = annotation["imgnum"]
        elif "page_index" in annotation:
            sequence = self.get_sequence(annotation["page_index"])
        start = annotation["span"]["start"]
        end = annotation["span"]["end"]
        self.add_triple(subject, rdf.type, bdo["EtextPage"])
        self.add_triple(subject, bdo["seqNum"], Literal(sequence, datatype=XSD.integer))
        self.add_triple(
            subject, bdo["sliceEndChar"], Literal(end, datatype=XSD.integer)
        )
        self.add_triple(
            subject, bdo["sliceStartChar"], Literal(start, datatype=XSD.integer)
        )
        self.add_triple(bdr[f"UT{volume_basename}"], bdo["eTextHasPage"], subject)

    @staticmethod
    def get_sequence(page_index):
        number = int(re.search(r"\d+", page_index).group())
        return number * 2 if page_index[-1] == "b" else (number * 2) - 1


    def add_chunks(self, baselname, volume_string, language, previous_i = 0, start_cc=0, end_cc=-1):
        chunk_indexes = self.get_chunk_indexes(volume_string, language, start_cc, end_cc)
        for i in range(0, len(chunk_indexes) - 1):
            self.set_etext_chunk(
                previous_i + i, chunk_indexes[i], chunk_indexes[i + 1], baselname, volume_string, language
            )
        return previous_i + len(chunk_indexes)


    def set_etext_chunks(self, baselname, baseinfo):
        volume_string = self.openpecha.get_base(baselname)
        default_language = "bo"
        if "default_language" in baseinfo:
            default_language = baseinfo["default_language"]
        elif self.openpecha.meta.default_language is not None:
            default_language = self.openpecha.meta.default_language
        llayer = self.openpecha.get_layer(baselname, LayerEnum.language)
        if llayer is None:
            self.add_chunks(baselname, volume_string, default_language)
        else:
            last_index = 0
            ci = 0
            # for our purpose it's crucial that the annotations are returned sorted:
            sorted_annotations = sorted(llayer.annotations.values(), key=lambda x: x["span"]["start"])
            for annotation in sorted_annotations:
                if annotation["span"]["start"] > last_index:
                    ci = self.add_chunks(baselname, volume_string, default_language, ci, last_index, annotation["span"]["start"])
                ci = self.add_chunks(baselname, volume_string, annotation["language"], ci, annotation["span"]["start"], annotation["span"]["end"])
                last_index = annotation["span"]["end"]
            len_vol_str = len(volume_string)
            if last_index < len_vol_str:
                self.add_chunks(baselname, volume_string, default_language, ci, last_index, len_vol_str)

    def set_etext_chunk(self, i, start_char, end_char, baselname, volume_string, language):
        volume_basename = f"{self.lname}_{baselname}"
        etext = f"UT{volume_basename}"
        subject = bdr[f"UT{volume_basename}_{int(i):05}"]
        self.add_triple(subject, rdf.type, bdo["EtextChunk"])
        self.add_triple(
            subject,
            bdo["chunkContents"],
            Literal(volume_string[start_char:end_char], lang=language),
        )
        self.add_triple(
            subject, bdo["sliceEndChar"], Literal(end_char, datatype=XSD.integer)
        )
        self.add_triple(
            subject, bdo["sliceStartChar"], Literal(start_char, datatype=XSD.integer)
        )

        self.add_triple(bdr[etext], bdo["eTextHasChunk"], subject)

    def set_adm(self):
        subject_r = bda[self.lname]
        einst_r = bdr[self.lname]
        graph_r = bdg[self.lname]
        if hasattr(self.openpecha, 'get_rev'):
            rev = self.openpecha.get_rev()
            self.add_triple(subject_r, adm["gitRevision"], Literal(rev))

        self.add_triple(subject_r, rdf.type, adm["AdminData"])
        self.add_triple(subject_r, adm["adminAbout"], einst_r)
        self.add_triple(subject_r, adm["graphId"], graph_r)
        self.add_triple(subject_r, adm["syncAgent"], bdr["SAOPT"])
        self.add_triple(subject_r, adm["metadataLegal"], bda["LD_BDRC_CC0"])
        

    @staticmethod
    def get_chunk_indexes(string, language, start_char = 0, end_char = 0):
        chunker = None
        if language == "bo":
            chunker = TibetanEasyChunker(string, 1500, start_char, end_char)
        else:
            chunker = EnglishEasyChunker(string, 1500, start_char, end_char)
        indexes = chunker.get_chunks()

        return indexes

    """
    Getting details of the rdf
    """

    def print_rdf(self):
        print(self.lod_g.serialize(format="ttl").decode("utf-8"))




import json

import requests

from openpecha import config


def parse_work_url(work_url):
    """Extract work id from work url

    Args:
        work_url (str): work url

    Returns:
        str: bdrc work id
    """
    work_id = ""
    if work_url:
        work_url_parts = work_url.split("/")
    work_id = work_url_parts[-1]
    return work_id


def get_bdrc_work_id(bdrc_inst_id):
    """Return bdrc work id to which the given bdrc instance id belongs to

    Args:
        bdrc_inst_id (str): bdrc instance id

    Returns:
        str: bdrc work id
    """
    bdrc_work_id = ""
    instance_response = requests.get(
        f"https://purl.bdrc.io/query/graph/OP_info?R_RES=bdr:{bdrc_inst_id}&format=json"
    )
    instance = json.loads(instance_response.content)
    instance_resource = instance.get(f"http://purl.bdrc.io/resource/{bdrc_inst_id}", {})
    instance_work_meta = instance_resource.get(
        "http://purl.bdrc.io/ontology/core/instanceOf", []
    )
    if instance_work_meta:
        work_url = instance_work_meta[0].get("value", "")
        bdrc_work_id = parse_work_url(work_url)
    else:
        print("Instance is not mapped to a work yet!!!")
    return bdrc_work_id


def parse_op2bdrc_work_id(op2bdrc_work_id):
    """Return op work id and bdrc work id

    Args:
        op2bdrc_work_id (str): openpecha work id and bdrc work id

    Returns:
        list: bdrc work id and op work id
    """
    op2bdrc_work_id = str(op2bdrc_work_id)
    work_ids = op2bdrc_work_id.split(",")
    op_work_id = work_ids[0].replace("b'", "")
    bdrc_work_id = work_ids[1].replace("'", "")
    return [bdrc_work_id, op_work_id]


def get_op_work_from_bdrc_work_id(bdrc_work_id, op2bdrc_mapping):
    """Return op work id from bdrc work id

    Args:
        bdrc_work_id (str): bdrc work id
        op2bdrc_mapping (str): op work and bdrc work id mapping

    Returns:
        str: openpecha work id
    """
    op_work_id = ""
    op2bdrc_works = op2bdrc_mapping.splitlines()
    for op2bdrc_work in op2bdrc_works[1:]:
        work_ids = parse_op2bdrc_work_id(op2bdrc_work)
        if work_ids[0] == bdrc_work_id:
            return work_ids[1]
    return op_work_id


def get_op_work_id(bdrc_instance_id):
    """Return openpecha work id of the given bdrc instance id

    Args:
        bdrc_inst_id (str): bdrc instance id

    Returns:
        str: openpecha work id
    """
    op_work_id = ""
    bdrc_work_id = get_bdrc_work_id(bdrc_instance_id)
    op2bdrc_work_mapping = requests.get(config.OP2BDRC_WORK_MAPPING_URL).content
    if bdrc_work_id:
        op_work_id = get_op_work_from_bdrc_work_id(bdrc_work_id, op2bdrc_work_mapping)
    return op_work_id



from botok.tokenizers.sentencetokenizer import sentence_tokenizer
from botok.tokenizers.wordtokenizer import WordTokenizer


def serialize_sentence(sentence):
    new_line = ""
    for token in sentence:
        new_line += f"{token.text}"
    new_line = new_line.strip()
    return new_line


def get_sentence_segments(text):
    wt = WordTokenizer()
    tokens = wt.tokenize(text, split_affixes=True)
    sentences = sentence_tokenizer(tokens)
    new_bo_txt = ""
    for sentence in sentences:
        new_bo_txt += serialize_sentence(sentence["tokens"]) + "\n"
    return new_bo_txt



from .docx import DocxSerializer
from .editor import EditorSerializer
from .epub import EpubSerializer
from .hfml import HFMLSerializer
from .pedurma import PedurmaSerializer
from .rdf import BUDARDFSerializer



from pathlib import Path

from openpecha.core.layer import Layer, LayerEnum
from openpecha.core.metadata import InitialCreationType, PechaMetadata
from openpecha.core.pecha import OpenPechaFS

from .. import config


class EmptyEbook:
    def __init__(self, output_path=config.PECHAS_PATH, metadata={}, assets={}) -> None:
        self.output_path = Path(output_path if output_path else config.PECHAS_PATH)
        self.metadata = metadata
        self.meta_fn = None
        self.pecha_path = None
        self.assets = assets
        self.layers = [
            LayerEnum.book_title,
            LayerEnum.sub_title,
            LayerEnum.book_number,
            LayerEnum.author,
            LayerEnum.chapter,
            LayerEnum.citation,
            LayerEnum.tsawa,
            LayerEnum.sabche,
            LayerEnum.yigchung,
        ]

    def get_dummy_layers(self):
        layers = {}
        for layer in self.layers:
            layers[layer] = Layer(
                annotation_type=layer, revision="00001", annotations={}
            )
        return layers

    def create_opf(self, text, id_):
        openpecha = OpenPechaFS(
            base={"v001": text},
            layers={"v001": self.get_dummy_layers()},
            metadata=PechaMetadata(
                initial_creation_type=InitialCreationType.ebook,
                source_metadata=self.metadata,
            ),
            assets=self.assets,
        )

        openpecha.save(output_path=self.output_path)
        self.meta_fn = openpecha.meta_fn
        self.pecha_path = openpecha.opf_path.parent



from pathlib import Path

from git import Repo

from openpecha.github_utils import commit, create_orphan_branch
from openpecha.utils import download_pecha, load_yaml

from .exporter.bitext import BitextExporter
from .exporter.po import PoExporter


class Alignment:
    def __init__(self, path=None, id: str = None):
        if not path:
            self.id = id
            self.alignment_repo_path = download_pecha(self.id)
        else:
            self.alignment_repo_path = Path(path)
            self.id = self.alignment_repo_path.stem
        self.metadata = None

    @classmethod
    def from_path(cls, path):
        if not Path(path).is_dir():
            raise FileNotFoundError(f"Alignment repo {path}")
        return cls(path=path)

    @property
    def meta_path(self):
        return self.alignment_repo_path / f"{self.id}.opa" / "meta.yml"

    def _load_metadata(self):
        repo = Repo(self.alignment_repo_path)
        pre_active_branch = repo.active_branch.name
        repo.git.checkout("master")
        metadata = load_yaml(self.meta_path)
        repo.git.checkout(pre_active_branch)
        return metadata

    @property
    def title(self):
        if self.metadata:
            return self.metadata["title"]

        self.metadata = self._load_metadata()
        return self.metadata["title"]

    @property
    def alignment_path(self):
        return self.alignment_repo_path / f"{self.id}.opa" / "Alignment.yml"

    def create(self):
        pass

    def create_po_view(self):
        exporter = PoExporter(self.alignment_path)
        repo = Repo(self.alignment_repo_path)
        if not self.is_po_created(repo):
            create_orphan_branch(
                self.alignment_repo_path, "po", parent_branch="master", type_="opa"
            )
        else:
            repo.git.checkout("po")
        exporter.export(self.alignment_repo_path)
        commit(self.alignment_repo_path, "po file added", not_includes=[], branch="po")
        return self.alignment_repo_path

    def is_po_created(self, repo):
        if "po" in repo.branches:
            return True
        else:
            return False

    def get_po_view(self):
        po_views = {}
        exporter = PoExporter(self.alignment_path)
        repo = Repo(self.alignment_repo_path)
        seg_srcs = exporter.alignment.get("segment_sources", {})
        if seg_srcs:
            if not self.is_po_created(repo):
                self.create_po_view()
            else:
                repo.git.checkout("po")
            for pecha_id, pecha_info in seg_srcs.items():
                pecha_type = pecha_info.get("relation", "")
                pecha_lang = pecha_info.get("lang", "")
                if pecha_type:
                    po_views[pecha_type] = {
                        "path": self.alignment_repo_path / f"{pecha_lang}.po",
                        "lang": pecha_lang,
                    }
        return po_views

    def create_bitext_view(self):
        exporter = BitextExporter(self.alignment_path)
        create_orphan_branch(
            self.alignment_repo_path, "bitext", parent_branch="master", type_="opa"
        )
        exporter.export(self.alignment_repo_path)
        commit(
            self.alignment_repo_path,
            "bitext file added",
            not_includes=[],
            branch="bitext",
        )
        return self.alignment_repo_path



from pathlib import Path
from uuid import uuid4

from openpecha import config
from openpecha.utils import dump_yaml, load_yaml


class TMXAlignment:
    def get_all_ids(self, annotations):
        curr_segment = {}
        final_segments = {}
        num = 1
        for uid, _ in annotations.items():
            curr_segment[num] = {"segment_id": uid}
            final_segments.update(curr_segment)
            curr_segment = {}
            num += 1
        return final_segments, num

    def get_segment_of_source(self, source_pecha_path):

        source_segment_yml = load_yaml(
            Path(
                source_pecha_path
                / f"{source_pecha_path.stem}.opf"
                / "layers/0001/Segment.yml"
            )
        )
        source_segment_ids, nums = self.get_all_ids(source_segment_yml["annotations"])
        segment = {}
        curr_seg = {}
        for num in range(1, nums):
            curr_seg[uuid4().hex] = {
                f"{source_pecha_path.stem}": source_segment_ids[num]["segment_id"]
            }
            segment.update(curr_seg)
            curr_seg = {}
        return segment

    def get_segment_pairs(self, source_pecha_path, target_pecha_path):

        source_segment_yml = load_yaml(
            Path(
                source_pecha_path
                / f"{source_pecha_path.stem}.opf"
                / "layers/0001/Segment.yml"
            )
        )
        target_segment_yml = load_yaml(
            Path(
                target_pecha_path
                / f"{target_pecha_path.stem}.opf"
                / "layers/0001/Segment.yml"
            )
        )

        source_segment_ids, s_nums = self.get_all_ids(source_segment_yml["annotations"])
        target_segment_ids, t_nums = self.get_all_ids(target_segment_yml["annotations"])

        curr_pair = {}
        segment_pairs = {}
        if s_nums == t_nums:
            for num in range(1, s_nums):
                curr_pair[uuid4().hex] = {
                    f"{source_pecha_path.stem}": source_segment_ids[num]["segment_id"],
                    f"{target_pecha_path.stem}": target_segment_ids[num]["segment_id"],
                }
                segment_pairs.update(curr_pair)
                curr_pair = {}
        return segment_pairs

    def update_alignment_repo(self, alignment_path, target_pecha_path, annotation_map):
        curr_pair = {}
        new_segment_pairs = {}
        target_segment = {}
        segment_source = {}
        target_pecha_id = target_pecha_path.stem
        alignment_id = alignment_path.stem
        target_meta_yml = load_yaml(
            target_pecha_path / f"{target_pecha_id}.opf" / "meta.yml"
        )
        alignment_yml = load_yaml(
            alignment_path / f"{alignment_id}.opa" / "Alignment.yml"
        )
        segment_source = alignment_yml["segment_sources"]
        for uid, info in segment_source.items():
            if info["relation"] == "source":
                source_pecha_id = uid
        target_segment[target_pecha_id] = {
            "type": target_meta_yml.get("origin_type", None),
            "relation": "target",
            "lang": target_meta_yml.get("source_metadata", {}).get("langauge", ""),
        }
        segment_source.update(target_segment)
        segment_pairs = alignment_yml["segment_pairs"]
        if annotation_map:
            for alignment_id, annotation_id in annotation_map.items():
                segment_pair = segment_pairs[alignment_id]
                curr_pair[alignment_id] = {
                    f"{source_pecha_id}": segment_pair[source_pecha_id],
                    f"{target_pecha_id}": annotation_id["annotation_id"],
                }
                new_segment_pairs.update(curr_pair)
                curr_pair = {}
        alignment_yml["segment_sources"] = segment_source
        alignment_yml["segment_pairs"] = new_segment_pairs
        self.write_alignment_repo(alignment_path, alignment_yml)

    def create_readme_for_opa(self, alignment_id, meta_yml):
        source_metadata = meta_yml["source_metadata"]
        type = meta_yml["type"]
        alignment = f"|Alignment id | {alignment_id}"
        Table = "| --- | --- "
        Title = f"|Title | {meta_yml['title']} "
        type = f"|Type | {type}"
        if source_metadata:
            srclang = f"|Source Language | {meta_yml['source_metadata']['srclang']}"
            tarlang = f"|Target Language | {meta_yml['source_metadata']['adminlang']}"
        else:
            srclang = None
            tarlang = None
        readme = f"{alignment}\n{Table}\n{Title}\n{type}\n{srclang}\n{tarlang}"
        return readme

    def _mkdir(self, path: Path):
        path.mkdir(parents=True, exist_ok=True)
        return path

    def write_alignment_repo(self, alignment_path, alignment_yml, meta_yml=None):
        alignment_opa_path = alignment_path / f"{alignment_path.stem}.opa"
        self._mkdir(alignment_opa_path)
        alignment_yml_path = alignment_opa_path / "Alignment.yml"
        meta_path = alignment_opa_path / "meta.yml"
        dump_yaml(alignment_yml, alignment_yml_path)
        if meta_yml:
            dump_yaml(meta_yml, meta_path)

    def create_alignment_meta(self, alignment_id, title, source_metadata, origin_type):
        metadata = {
            "id": alignment_id,
            "title": title,
            "type": origin_type,
            "source_metadata": source_metadata,
        }
        return metadata

    def create_alignment_yml(
        self, source_pecha_path, target_pecha_path, src_lang, tar_lang, origin_type
    ):
        alignment_id = uuid4().hex
        if target_pecha_path:
            segment_pairs = self.get_segment_pairs(source_pecha_path, target_pecha_path)
            alignment = {
                "segment_sources": {
                    f"{source_pecha_path.stem}": {
                        "type": origin_type,
                        "relation": "source",
                        "lang": src_lang,
                    },
                    f"{target_pecha_path.stem}": {
                        "type": origin_type,
                        "relation": "target",
                        "lang": tar_lang,
                    },
                },
                "segment_pairs": segment_pairs,
            }
        else:
            segment = self.get_segment_of_source(source_pecha_path)
            alignment = {
                "segment_sources": {
                    f"{source_pecha_path.stem}": {
                        "type": origin_type,
                        "relation": "source",
                        "lang": src_lang,
                    }
                },
                "segment_pairs": segment,
            }
        return alignment_id, alignment

    def create_alignment_repo(
        self,
        source_pecha_path,
        target_pecha_path=None,
        title=None,
        source_metadata=None,
        origin_type="translation",
    ):
        if source_metadata:
            src_lang = source_metadata.get("srclang", "")
            tar_lang = source_metadata.get("adminlang", "")
        else:
            src_lang = None
            tar_lang = None

        alignment_id, alignment_yml = self.create_alignment_yml(
            source_pecha_path, target_pecha_path, src_lang, tar_lang, origin_type
        )

        alignment_path = config.PECHAS_PATH / alignment_id / f"{alignment_id}.opa"

        meta_yml = self.create_alignment_meta(
            alignment_id, title, source_metadata, origin_type
        )

        self.write_alignment_repo(alignment_path.parent, alignment_yml, meta_yml)

        readme = self.create_readme_for_opa(alignment_id, meta_yml)
        (alignment_path.parent / "readme.md").write_text(readme, encoding="utf-8")
        return alignment_path.parent



from collections import defaultdict
from typing import Dict

from bs4 import BeautifulSoup
from pydantic.tools import T

from openpecha.core.annotation import AnnBase, Span
from openpecha.core.layer import Layer, LayerEnum
from openpecha.serializers.epub import TsadraTemplateCSSClasses


class EditorParser:
    def __init__(self):
        self._reset()

    def _reset(self):
        self.layers: Dict[str, Dict[str, Layer]] = defaultdict(dict)
        self.base: Dict[str, str] = {}
        self.last_base_char_idx = -1

    def _add_base(self, root, base_name):
        self.base[base_name] = root.text

    def _get_content_span(self, text):
        start = self.last_base_char_idx + 1
        end = start + len(text) - 1
        self.last_base_char_idx = end
        return start, end

    def _get_ann(self, tag, ann_class, extra_attrs):
        start, end = self._get_content_span(tag.text)
        span = Span(start=start, end=end)
        return ann_class(span=span, **extra_attrs), tag["id"]

    @staticmethod
    def _get_empty_layer(layerEnum):
        return Layer(annotation_type=layerEnum, revision="0000", annotations={})

    def _get_layer(self, base_name, layerEnum):
        layer = self.layers[base_name].get(layerEnum, self._get_empty_layer(layerEnum))
        self.layers[base_name][layerEnum] = layer
        return layer

    def _add_ann(self, base_name, layerEnum, tag, ann_class=AnnBase, extra_attrs={}):
        ann, id_ = self._get_ann(tag, ann_class, extra_attrs)
        layer = self._get_layer(base_name, layerEnum)
        layer.annotations[id_] = ann

    def __handle_non_ann(self, text):
        start, end = self._get_content_span(text)

    def _parse_p_tag(self, base_name, p):
        for child in p.children:
            if isinstance(child, str):
                self.__handle_non_ann(child)
                continue

            tag_class = child["class"]
            if "book-title" in tag_class:
                self._add_ann(base_name, LayerEnum.book_title, child)
            elif "sub-title" in tag_class:
                self._add_ann(base_name, LayerEnum.sub_title, child)
            elif "book-number" in tag_class:
                self._add_ann(base_name, LayerEnum.book_number, child)
            elif "author" in tag_class:
                self._add_ann(base_name, LayerEnum.author, child)
            elif "chapter" in tag_class:
                self._add_ann(base_name, LayerEnum.chapter, child)
            elif "sabche" in tag_class:
                self._add_ann(base_name, LayerEnum.sabche, child)
            elif "yigchung" in tag_class:
                self._add_ann(base_name, LayerEnum.yigchung, child)
            elif "citation" in tag_class:
                self._add_ann(
                    base_name,
                    LayerEnum.citation,
                    child,
                    extra_attrs={
                        "metadata": {
                            "css_class_name": TsadraTemplateCSSClasses.citation_inline.value
                        }
                    },
                )
            elif "citation-verse" in tag_class:
                self._add_ann(
                    base_name,
                    LayerEnum.citation,
                    child,
                    extra_attrs={
                        "metadata": {
                            "css_class_name": TsadraTemplateCSSClasses.citation_verse.value
                        }
                    },
                )
            elif "citation-prose" in tag_class:
                self._add_ann(
                    base_name,
                    LayerEnum.citation,
                    child,
                    extra_attrs={
                        "metadata": {
                            "css_class_name": TsadraTemplateCSSClasses.citation_prose.value
                        }
                    },
                )
            elif "root-text" in tag_class:
                self._add_ann(
                    base_name,
                    LayerEnum.tsawa,
                    child,
                    extra_attrs={
                        "metadata": {
                            "css_class_name": TsadraTemplateCSSClasses.tsawa_inline.value
                        }
                    },
                )
            elif "root-text-verse" in tag_class:
                self._add_ann(
                    base_name,
                    LayerEnum.tsawa,
                    child,
                    extra_attrs={
                        "metadata": {
                            "css_class_name": TsadraTemplateCSSClasses.tsawa_verse.value
                        }
                    },
                )

        # newline at the end of every p tag
        self.last_base_char_idx += 1

    def _group_verse(self, base_name, layer_name):
        """Group same typed of annotation in consicutive order."""

        def _create_ann(start, end, is_verse):
            pass

        if layer_name not in self.layers[base_name]:
            return

        layer = self.layers[base_name][layer_name]
        verse_grouped_anns = {}

        anns = sorted(layer.annotations.items(), key=lambda ann: ann[1].span.start)

        verse_ended = True
        start_verse_ann_id = None
        prev_verse_ann = None
        for i, (ann_id, ann) in enumerate(anns):
            if "verse" in ann.metadata["css_class_name"]:
                if verse_ended:
                    verse_grouped_anns[ann_id] = ann
                    start_verse_ann_id = ann_id
                    verse_ended = False
                else:
                    prev_verse_ann = ann

                # if last ann is verse
                if i+1 == len(anns):
                    verse_grouped_anns[start_verse_ann_id].span.end = prev_verse_ann.span.end
            else:
                verse_grouped_anns[ann_id] = ann
                if not verse_ended:
                    verse_grouped_anns[start_verse_ann_id].span.end = prev_verse_ann.span.end
                    verse_ended = True

        layer.annotations = verse_grouped_anns


    def parse(self, base_name, html, group_verse=True):
        self._reset()

        root = BeautifulSoup(html, "html.parser")
        self._add_base(root, base_name)

        for p in root.find_all("p"):
            self._parse_p_tag(base_name, p)

        if group_verse:
            self._group_verse(base_name, LayerEnum("Tsawa"))
            self._group_verse(base_name, LayerEnum("Citation"))



import json
from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, validator

from .annotations import *
from .ids import get_uuid
from .metadata import PechaMetadata


class LayerEnum(Enum):
    index = "index"

    book_title = "BookTitle"
    sub_title = "SubTitle"
    book_number = "BookNumber"
    poti_title = "PotiTitle"
    author = "Author"
    chapter = "Chapter"

    topic = "Text"
    sub_topic = "SubText"

    pagination = "Pagination"
    language = "Language"
    citation = "Citation"
    correction = "Correction"
    error_candidate = "ErrorCandidate"
    peydurma = "Peydurma"
    sabche = "Sabche"
    tsawa = "Tsawa"
    yigchung = "Yigchung"
    archaic = "Archaic"
    durchen = "Durchen"
    footnote = "Footnote"
    segment = "Segment"
    ocr_confidence = "OCRConfidence"


def _get_annotation_class(layer_name: LayerEnum):
    """Maps LayerEnum to Annotation class"""
    if layer_name == LayerEnum.book_title:
        return BaseAnnotation
    elif layer_name == LayerEnum.sub_title:
        return BaseAnnotation
    elif layer_name == LayerEnum.book_number:
        return BaseAnnotation
    elif layer_name == LayerEnum.poti_title:
        return BaseAnnotation
    elif layer_name == LayerEnum.author:
        return BaseAnnotation
    elif layer_name == LayerEnum.chapter:
        return BaseAnnotation
    elif layer_name == LayerEnum.topic:
        return BaseAnnotation
    elif layer_name == LayerEnum.sub_topic:
        return BaseAnnotation
    elif layer_name == LayerEnum.pagination:
        return Pagination
    elif layer_name == LayerEnum.language:
        return Language
    elif layer_name == LayerEnum.citation:
        return Citation
    elif layer_name == LayerEnum.correction:
        return Correction
    elif layer_name == LayerEnum.error_candidate:
        return ErrorCandidate
    elif layer_name == LayerEnum.peydurma:
        return Pedurma
    elif layer_name == LayerEnum.sabche:
        return Sabche
    elif layer_name == LayerEnum.tsawa:
        return Tsawa
    elif layer_name == LayerEnum.yigchung:
        return Yigchung
    elif layer_name == LayerEnum.archaic:
        return Archaic
    elif layer_name == LayerEnum.durchen:
        return Durchen
    elif layer_name == LayerEnum.footnote:
        return Footnote
    elif layer_name == LayerEnum.segment:
        return Segment
    elif layer_name == LayerEnum.ocr_confidence:
        return OCRConfidence
    else:
        return BaseAnnotation

class Layer(BaseModel):
    id: str = None
    annotation_type: LayerEnum
    revision: str = "00001"
    annotations: Dict = {}

    @validator("id", pre=True, always=True)
    def set_id(cls, v):
        return v or get_uuid()

    @validator("revision")
    def revision_must_int_parsible(cls, v):
        assert v.isdigit(), "must integer parsible like `00002`"
        return v

    def bump_revision(self):
        self.revision = f"{int(self.revision)+1:05}"

    def reset(self):
        self.revision = "00001"
        self.annotations = {}

    def get_annotations(self):
        """Yield Annotation Objects"""
        for ann_id, ann_dict in self.annotations.items():
            ann_class = _get_annotation_class(self.annotation_type)
            ann = ann_class.parse_obj(ann_dict)
            yield ann_id, ann

    def get_annotation(self, annotation_id: str) -> Optional[BaseAnnotation]:
        """Retrieve annotation of id `annotation_id`"""
        ann_dict = self.annotations.get(annotation_id)
        if not ann_dict:
            return
        ann_class = _get_annotation_class(self.annotation_type)
        ann = ann_class.parse_obj(ann_dict)
        return ann

    def set_annotation(self, ann: BaseAnnotation, ann_id = None):
        """Add or Update annotation `ann` to the layer, returns the annotation id"""
        ann_id = ann_id if ann_id is not None else get_uuid()
        self.annotations[ann_id] = json.loads(ann.json())
        return ann_id

    def remove_annotation(self, annotation_id: str):
        """Delete annotaiton of `annotation_id` from the layer"""
        if annotation_id in self.annotations:
            del self.annotations[annotation_id]


class SpanINFO(BaseModel):
    text: str
    layers: Dict[LayerEnum, List[BaseAnnotation]]
    metadata: PechaMetadata


class OCRConfidenceLayer(Layer):
    confidence_threshold: float
    annotation_type: LayerEnum = LayerEnum.ocr_confidence


"""Utilities functions
"""
import gzip
import io
import shutil
from collections import defaultdict
from pathlib import Path
from typing import Dict

import yaml
from git import GIT_OK, Repo
from git.cmd import GitCommandError

from openpecha import config, storages
from openpecha.exceptions import PechaNotFound
from openpecha.github_utils import create_release
from openpecha.storages import GithubStorage, setup_auth_for_old_repo

INFO = "[INFO] {}"

# use yaml.CSafeLoader / if available but don't crash if it isn't
try:
    yaml_loader = yaml.CSafeLoader
except (ImportError, AttributeError):
    yaml_loader = yaml.SafeLoader

try:
    yaml_dumper = yaml.CSafeDumper
except (ImportError, AttributeError):
    yaml_dumper = yaml.SafeDumper

def gzip_str(string_):
    # taken from https://gist.github.com/Garrett-R/dc6f08fc1eab63f94d2cbb89cb61c33d
    out = io.BytesIO()

    with gzip.GzipFile(fileobj=out, mode="w") as fo:
        fo.write(string_.encode())

    bytes_obj = out.getvalue()
    return bytes_obj

def _mkdir(path):
    if path.is_dir():
        return path
    path.mkdir(exist_ok=True, parents=True)
    return path
    
def ocr_result_input(path):
    return path


def create_release_with_assets(path):
    asset_paths = []
    for asset_path in (path / "releases").iterdir():
        shutil.make_archive(asset_path, "zip", asset_path)
        asset_paths.append(f"{str(asset_path)}.zip")

    create_release(path.name, asset_paths=asset_paths)


class Vol2FnManager:
    def __init__(self, metadata):
        self.name = "vol2fn"
        self.vol_num = 0
        self.vol2fn = self._get_vol2fn(metadata)
        self.fn2vol = {fn: vol for vol, fn in self.vol2fn.items()}

    def _get_vol2fn(self, metadata):
        if self.name in metadata:
            return metadata[self.name]
        else:
            return defaultdict(dict)

    def get_fn(self, vol):
        return self.vol2fn.get(vol)

    def get_vol_id(self, fn):
        vol_id = self.fn2vol.get(fn)
        if vol_id:
            return vol_id
        else:
            self.vol_num += 1
            vol_id = f"v{self.vol_num:03}"
            self.vol2fn[vol_id] = fn
            return vol_id


def dump_yaml(data: Dict, output_fn: Path) -> Path:
    with output_fn.open("w", encoding="utf-8") as fn:
        yaml.dump(
            data,
            fn,
            default_flow_style=False,
            sort_keys=False,
            allow_unicode=True,
            Dumper=yaml_dumper,
        )
    return output_fn


def load_yaml(fn: Path) -> None:
    return yaml.load(fn.open(encoding="utf-8"), Loader=yaml_loader)


def load_yaml_str(s: str) -> None:
    return yaml.load(s, Loader=yaml_loader)

def _eval_branch(repo, branch):
    """return default branch as fallback branch."""
    if branch in repo.refs or f"origin/{branch}" in repo.refs:
        return branch
    elif "main" in repo.refs:
        return "main"
    else:
        return "master"


def download_pecha(pecha_id, out_path=None, needs_update=True, branch="main"):
    # clone the repo
    pecha_url = f"{config.GITHUB_ORG_URL}/{pecha_id}.git"
    if out_path:
        out_path = Path(out_path)
        out_path.mkdir(exist_ok=True, parents=True)
        pecha_path = out_path / pecha_id
    else:
        pecha_path = config.PECHAS_PATH / pecha_id

    if pecha_path.is_dir():  # if repo is already exits at local then try to pull
        repo = Repo(str(pecha_path))
        branch = _eval_branch(repo, branch)
        repo.git.checkout(branch)
        if needs_update:
            print(INFO.format(f"Updating {pecha_id} ..."))
            repo.git.pull("origin", branch)
    else:
        print(INFO.format(f"Downloading {pecha_id} ..."))
        try:
            Repo.clone_from(pecha_url, str(pecha_path))
        except GitCommandError:
            raise PechaNotFound(f"Pecha with id {pecha_id} doesn't exist")
        repo = Repo(str(pecha_path))
        branch = _eval_branch(repo, branch)
        repo.git.checkout(branch)

    # setup auth
    storage = GithubStorage()
    setup_auth_for_old_repo(repo, org=storage.org_name, token=storage.token)

    return pecha_path



import csv
import io
import json
import logging
import os
import shutil
import sys
import zipfile

import requests
from genericpath import exists
from tqdm import tqdm

from openpecha import config, utils

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(message)s",
    handlers=[logging.StreamHandler(sys.stdout)],
)


def get_github_token():
    token = os.environ.get("GITHUB_TOKEN")
    if token:
        return token
    github_keys_path = config.BASE_PATH / "keys" / "github.json"
    if github_keys_path.is_file():
        keys = json.load(github_keys_path.open())
        return keys["download"]

    url = "https://github.com/OpenPecha-dev/keys/releases/download/v1/keys.zip"
    r = requests.get(url)
    z = zipfile.ZipFile(io.BytesIO(r.content))
    z.extractall(path=str(config.BASE_PATH))
    keys = json.load(github_keys_path.open())
    return keys["download"]


def get_base_vol_list(pecha_repo, pecha_id):
    """Return list of base text file name present in pecha's opf

    Args:
        pecha_repo (repo): pecha repo object
        pecha_id (str): pecha id

    Returns:
        list: list of base text file name
    """
    base_vols = []
    contents = pecha_repo.get_contents(f"/{pecha_id}.opf/base")
    for content_file in contents:
        base_vols.append(content_file.name)
    return base_vols


def download_base_vols(output_path, pecha_id, base_vols, session):
    """Download and save pecha's base text in output path dir

    Args:
        output_path (Path): output path dir
        pecha_id (sttr): pecha id
        base_vols (list): list of base text in opf
    """
    for base_vol in base_vols:
        base_response = session.get(
            f"https://raw.githubusercontent.com/OpenPecha/{pecha_id}/master/{pecha_id}.opf/base/{base_vol}"
        )
        base_text = base_response.text
        pecha_dir_path = utils._mkdir((output_path / pecha_id))
        (pecha_dir_path / base_vol).write_text(base_text, encoding="utf-8")


def get_corpus_catalog(corpus_name, session):
    catalog_url = (
        f"https://github.com/OpenPecha/corpus_catalog/raw/main/data/{corpus_name}.csv"
    )
    r = session.get(catalog_url, stream=True)

    if r.status_code != requests.codes.OK:
        return ValueError(f"Corpus with name `{corpus_name}` doesn't exists")

    lines = (line.decode("utf-8") for line in r.iter_lines())
    csv_reader = csv.reader(lines)
    next(csv_reader, None)  # skip headers
    for row in csv_reader:
        yield row


def get_corpus_items_count(corpus_name, session):
    url = (
        f"https://github.com/OpenPecha/corpus_catalog/raw/main/data/{corpus_name}.count"
    )
    r = session.get(url)

    if r.status_code != requests.codes.OK:
        logging.info("Please contact developer to set corpus item count")

    return int(r.text.strip())


def get_request_session():
    session = requests.Session()
    token = get_github_token()
    session.headers["Authorization"] = f"token {token}"
    return session


def download_zip_file(url, pecha_path, session=None):
    if session:
        r = session.get(url, stream=True)
    else:
        r = requests.get(url, stream=True)

    if r.status_code != requests.codes.OK:
        return

    corpus_path = pecha_path.parent
    out_fn_zip = corpus_path / f"{pecha_path.name}.zip"
    with out_fn_zip.open("wb") as f:
        for chunk in r.iter_content(chunk_size=512):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)

    # unzip the file
    with zipfile.ZipFile(str(out_fn_zip), "r") as zip:
        zip.extractall(corpus_path)
        members = zip.namelist()

    out_fn_zip.unlink()
    unzipfile_path = corpus_path / members[0][:-1]
    shutil.move(str(unzipfile_path), str(pecha_path))


def download_corpus(
    corpus_name, output_path=utils._mkdir(config.BASE_PATH / "corpus"), replace=False
):
    """download corpus from openpecha

    Args:
        corpus_name (str): name of corpus on which list of pecha has been prepared in catalog repo of openpecha
        output_path (Path, optional): output path where corpus will be saved. Defaults to None.

    Return:
        path: output path
    """
    corpus_output_path = output_path / corpus_name
    corpus_output_path.mkdir(exist_ok=True, parents=True)
    session = get_request_session()
    corpus_catalog = get_corpus_catalog(corpus_name, session)
    corpus_items_count = get_corpus_items_count(corpus_name, session)
    logging.info(f"Downloading `{corpus_name}` corpus...")
    for corpus in tqdm(corpus_catalog, total=corpus_items_count):
        pecha_id, download_url, quality = corpus
        pecha_path = corpus_output_path / pecha_id
        if pecha_path.is_dir() and not replace:
            continue
        download_zip_file(download_url, pecha_path, session)


if __name__ == "__main__":
    download_corpus("literary_bo")



from .empty import EmptyEbook
from .formatter import BaseFormatter
from .google_vision import GoogleVisionFormatter
from .hfml import HFMLFormatter
from .tsadra import TsadraFormatter
from .pedurma import PedurmaFormatter



from datetime import datetime
from enum import Enum
from typing import Dict, Optional, Union

from pydantic import AnyHttpUrl, BaseModel, Extra, validator

from . import ids


class InitialCreationType(Enum):
    ocr = "ocr"
    ebook = "ebook"
    input = "input"


class CopyrightStatus(Enum):
    UNKNOWN = "Unknown"
    COPYRIGHTED = "In copyright"
    PUBLIC_DOMAIN = "Public domain"


class Copyright(BaseModel):
    status: CopyrightStatus = CopyrightStatus.UNKNOWN
    notice: Optional[str] = ""
    info_url: Optional[AnyHttpUrl] = None

    class Config:
        extra = Extra.forbid

Copyright_copyrighted = Copyright(
    status=CopyrightStatus.COPYRIGHTED,
    notice="In copyright by the original author or editor",
    info_url="http://rightsstatements.org/vocab/InC/1.0/",
)

Copyright_unknown = Copyright(
    status=CopyrightStatus.UNKNOWN,
    notice="Copyright Undertermined",
    info_url="http://rightsstatements.org/vocab/UND/1.0/",
)

Copyright_public_domain = Copyright(
    status=CopyrightStatus.PUBLIC_DOMAIN,
    notice="Public domain",
    info_url="https://creativecommons.org/publicdomain/mark/1.0/",
)

class LicenseType(Enum):
    # based on https://creativecommons.org/licenses/

    CC0 = "CC0"
    PUBLIC_DOMAIN_MARK = "Public Domain Mark"
    CC_BY = "CC BY"
    CC_BY_SA = "CC BY-SA"
    CC_BY_ND = "CC BY-ND"
    CC_BY_NC = "CC BY-NC"
    CC_BY_NC_SA = "CC BY-NC-SA"
    CC_BY_NC_ND = "CC BY-NC-ND"
    
    UNDER_COPYRIGHT = "under copyright"


class PechaMetadata(BaseModel):
    id: str = None
    legacy_id: Optional[str] = None
    ocr_import_info: Optional[Dict] = None
    default_language: str = None
    source: str = None
    source_file: str = None
    initial_creation_type: InitialCreationType
    imported: datetime = None
    last_modified: datetime = None
    parser: AnyHttpUrl = None
    source_metadata: Optional[Dict] = None  # place to dump any metadata from the source
    statistics: Optional[Dict] = None
    quality: Optional[Dict] = None
    bases: Optional[Dict[str, Dict]] = {}
    copyright: Copyright = None
    license: LicenseType = None

    @validator("imported", pre=True, always=True)
    def set_imported_date(cls, v):
        return v or datetime.now()

    @validator("last_modified", pre=True, always=True)
    def set_last_modified_date(cls, v):
        return v or datetime.now()

    @validator("copyright", pre=True, always=True)
    def set_copyright_info(cls, v):
        return v or Copyright()

    def update_last_modified_date(self):
        self.last_modified = datetime.now()

    # class Config:
    #     extra = Extra.forbid


class InitialPechaMetadata(PechaMetadata):
    bases: Dict = {}

    @validator("id", pre=True, always=True)
    def set_id(cls, v):
        return v or ids.get_initial_pecha_id()


class OpenPechaMetadata(PechaMetadata):
    @validator("id", pre=True, always=True)
    def set_id(cls, v):
        return v or ids.get_open_pecha_id()


class DiplomaticPechaMetadata(PechaMetadata):
    @validator("id", pre=True, always=True)
    def set_id(cls, v):
        return v or ids.get_diplomatic_id()



import re
from pathlib import Path
from uuid import uuid4

import polib
from create_opf import create_readme, get_segment_annotation

from openpecha import config
from openpecha.core.annotation import AnnBase, Span
from openpecha.core.layer import Layer, LayerEnum
from openpecha.core.pecha import OpenPechaFS
from openpecha.utils import dump_yaml, load_yaml

from . import TMXAlignment


def read_po(path):
    return polib.pofile(path)


def get_id_and_string(po_file):
    po_dict = {}
    curr_po = {}
    for entry in po_file:
        id = entry.msgid
        string = entry.msgstr
        text = re.sub(r"\n", " ", string)
        curr_po[id] = {text}
        po_dict.update(curr_po)
        curr_po = {}
    return po_dict


def get_segment_layer_base_and_mapping(po_dict):
    segment_annotations = {}
    char_walker = 0
    base_text = ""
    annotation_map = {}
    curr_dict = {}
    for uid, text in po_dict.items():
        segment_annotation, end = get_segment_annotation(text, char_walker)
        segment_annotations.update(segment_annotation)
        char_walker = end + 2
        annotation_id = list(segment_annotation.keys())[0]
        base_text += list(text)[0] + "\n"
        curr_dict[uid] = {annotation_id}
        annotation_map.update(curr_dict)
        curr_dict = {}

    segment_layer = Layer(
        annotation_type=LayerEnum.segment, annotations=segment_annotations
    )
    return segment_layer, base_text, annotation_map


def create_opf_for_translated(
    po_dict, pecha_id, opf_path, title=None, origin_type="translation", lan=None
):
    opf = OpenPechaFS(path=opf_path)
    segment_layer, base_text, annotation_map = get_segment_layer_base_and_mapping(
        po_dict
    )
    layers = {"0001": {LayerEnum.segment: segment_layer}}
    bases = {"0001": base_text}
    metadata = {
        "id": pecha_id,
        "initial_creation_type": "TMX",
        "origin_type": origin_type,
        "source_metadata": {
            "id": "",
            "title": title,
            "langauge": lan,
            "author": "",
            "volume": {},
        },
    }
    opf.layers = layers
    opf.base = bases
    opf.save_base()
    opf.save_layers()
    meta_fn = Path(f"{opf_path}/meta.yml")
    dump_yaml(metadata, meta_fn)
    return annotation_map


def get_opf_path_and_annotation_map(po_zip, title, lan):
    if po_zip:
        pecha_id = uuid4().hex
        opf_path = config.PECHAS_PATH / pecha_id / f"{pecha_id}.opf/"
        annotation_map = create_opf_for_translated(
            po_zip, pecha_id, opf_path, title, lan=lan
        )
        readme = create_readme(pecha_id)
        Path(config.PECHAS_PATH / pecha_id / "readme.md").write_text(
            readme, encoding="utf-8"
        )
    return opf_path, annotation_map


def update_alignment_from_po(po_path):
    po_file = read_po(po_path)
    po_zip = get_id_and_string(po_file)
    lan = "bo"
    title = ""
    pecha_id, annotation_map = get_opf_path_and_annotation_map(po_zip, title, lan)
    alignment_id = ""
    alignment_repo = config.PECHAS_PATH / alignment_id / f"{alignment_id}.opa"
    obj = TMXAlignment
    obj.update_alignment_repo(
        alignment_repo=alignment_repo,
        target_pecha_id=pecha_id,
        annotation_map=annotation_map,
        lan=lan,
    )



# OpenPecha Toolkit: Corpus Quality Utility Functions
#
# Copyright (C) 2019-2022 NLTK Project
# Author: Edward Loper <edloper@gmail.com>
# URL: <https://www.nltk.org/>
# For license information, see LICENSE.TXT

from typing import Dict

from botok import WordTokenizer
from botok.vars import WordMarkers


class NonWordsCounter:
    """
    Class to count non words and it's ratio
    """

    def __init__(
        self,
        text: str = None,
        tokenizer: WordTokenizer = None,
        total_words: int = 0,
        total_non_words: int = 0,
        empty=False,
    ):
        self.total_words: int = total_words
        self.total_non_words: int = total_non_words
        if not total_non_words and not empty:
            self.count(text, tokenizer)

    def count(self, text, tokenizer):
        if not text:
            raise ValueError("required text input")
        tokenizer = tokenizer if tokenizer else WordTokenizer()
        tokens = tokenizer.tokenize(text)
        self.total_words = len(tokens)
        for token in tokens:
            if token.pos in [WordMarkers.NON_WORD.name, WordMarkers.NO_POS.name]:
                self.total_non_words += 1

    @property
    def non_word_ratio(self):
        return round(self.total_non_words / self.total_words, 2)

    def __add__(self, other):
        return NonWordsCounter(
            total_words=self.total_words + other.total_words,
            total_non_words=self.total_non_words + other.total_non_words,
        )



"""catalog module contains all the functionalities necessary for managin
the catalog. Functonalities includes:
    - Creating opfs from input text
    - Assiging ID to the new opf
    - Updating the catalog with new opfs

"""

import yaml

from openpecha.core.ids import get_initial_pecha_id
from openpecha.github_utils import create_readme, github_publish
from openpecha.storages import GithubStorage, Storage
from openpecha.utils import create_release_with_assets, ocr_result_input

buildin_pipes = {
    "input": {"ocr_result_input": ocr_result_input},
    "release": {"create_release_with_assets": create_release_with_assets},
}


class CatalogManager:
    """Manages the catalog"""

    def __init__(
        self,
        pipes=None,
        formatter=None,
        layers=[],
        not_include_files=["releases"],
        storage: Storage = None,
    ):
        self.repo_name = "catalog"
        self.batch_path = "data/batch.csv"
        self.batch = []
        self.formatter = formatter
        self.layers = layers
        self.not_include_files = not_include_files
        self.pipes = pipes if pipes else buildin_pipes
        self.storage = storage if storage else GithubStorage()

    def _add_id_url(self, row):
        id = row[0]
        row[0] = f"[{id}](https://github.com/{self.storage.org_name}/{id})"
        return row

    def update(self):
        """Updates the catalog csv to have new opf-pechas metadata"""
        content = (
            "\n".join([",".join(row) for row in map(self._add_id_url, self.batch)])
            + "\n"
        )
        self.storage.add_file(
            dir_name=self.repo_name,
            path=self.batch_path,
            content=content,
            message="update with new batch",
        )
        print("[INFO] Updated the catalog")

        # reset the batch
        self.batch = []

    def _get_catalog_metadata(self, meta_fn):
        metadata = yaml.safe_load(meta_fn.open())
        catalog_metadata = [
            metadata["id"].split(":")[-1],
            metadata["source_metadata"].get("title", ""),
            metadata["source_metadata"].get("subtitle", ""),
            metadata["source_metadata"].get("author", ""),
            metadata["source_metadata"].get("id", ""),
        ]
        self.batch.append(catalog_metadata)
        create_readme(metadata["source_metadata"], self.formatter.pecha_path)

    def format_and_publish(self, path):
        """Convert input pecha to opf-pecha with id assigined"""
        self.formatter.create_opf(path, get_initial_pecha_id())
        self._get_catalog_metadata(self.formatter.meta_fn)
        github_publish(
            self.formatter.pecha_path,
            not_includes=self.not_include_files,
            layers=self.layers,
            org=self.storage.org_name,
            token=self.storage.token,
        )
        return self.formatter.pecha_path

    def add_ocr_item(self, path):
        self._process(path, "ocr_result_input", "create_release_with_assets")

    def add_hfml_item(self, path):
        self._process(path, "ocr_result_input")

    def add_empty_item(self, text):
        self._process(text, "ocr_result_input")

    def _process(self, path, input_method, release_method=None):
        print("[INFO] Getting input")
        raw_pecha_path = self.pipes["input"][input_method](path)
        print("[INFO] Convert Pecha to OPF")
        opf_pecha_path = self.format_and_publish(raw_pecha_path)
        if release_method:
            print("[INFO] Release OPF pecha")
            self.pipes["release"][release_method](opf_pecha_path)



import logging
import shutil
import traceback
from pathlib import Path

import click
from git import Repo
from git.cmd import GitCommandError
from tqdm import tqdm

import openpecha
from openpecha import config
from openpecha.alignment.tmx import create_opf as alignment
from openpecha.blupdate import PechaBaseUpdate
from openpecha.catalog import config as catalog_config
from openpecha.catalog.filter import is_text_good_quality
from openpecha.catalog.storage import GithubBucket
from openpecha.core.pecha import OpenPechaFS
from openpecha.corpus.quality import NonWordsCounter
from openpecha.exceptions import PechaNotFound
from openpecha.formatters import GoogleOCRFormatter, HFMLFormatter
from openpecha.serializers import EpubSerializer, HFMLSerializer
from openpecha.storages import GithubStorage, setup_auth_for_old_repo

OP_PATH = config.BASE_PATH
config = {
    # Github
    "OP_CATALOG_URL": "https://raw.githubusercontent.com/OpenPoti/openpecha-catalog/master/data/catalog.csv",
    "OP_ORG": "https://github.com/OpenPecha",
    # Local
    "OP_DATA_PATH": (OP_PATH / "data").resolve(),
    "OP_PECHAS_PATH": config.PECHAS_PATH.resolve(),
    "OP_CATALOG_PATH": (OP_PATH / "data" / "catalog.csv").resolve(),
    "CONFIG_PATH": (OP_PATH / "config").resolve(),
    "DATA_CONFIG_PATH": (OP_PATH / "config" / "data_config").resolve(),
}

ERROR = "[ERROR] {}"
INFO = "[INFO] {}"


def get_pecha_id(n):
    return f"P{int(n):06}"


@click.group()
@click.version_option(version=openpecha.__version__)
def cli():
    pass


def create_config_dirs():
    config["OP_DATA_PATH"].mkdir(parents=True, exist_ok=True)
    config["CONFIG_PATH"].mkdir(parents=True, exist_ok=True)


def _eval_branch(repo, branch):
    """return default branch as fallback branch."""
    if branch in repo.refs or f"origin/{branch}" in repo.refs:
        return branch
    elif "main" in repo.refs:
        return "main"
    else:
        return "master"


def download_pecha(pecha_id, out_path=None, needs_update=True, branch="main"):
    # clone the repo
    pecha_url = f"{config['OP_ORG']}/{pecha_id}.git"
    if out_path:
        out_path = Path(out_path)
        out_path.mkdir(exist_ok=True, parents=True)
        pecha_path = out_path / pecha_id
    else:
        pecha_path = config["OP_PECHAS_PATH"] / pecha_id

    if pecha_path.is_dir():  # if repo is already exits at local then try to pull
        repo = Repo(str(pecha_path))
        branch = _eval_branch(repo, branch)
        repo.git.checkout(branch)
        if needs_update:
            click.echo(INFO.format(f"Updating {pecha_id} ..."))
            repo.git.pull("origin", branch)
    else:
        click.echo(INFO.format(f"Downloading {pecha_id} ..."))
        try:
            Repo.clone_from(pecha_url, str(pecha_path))
        except GitCommandError:
            raise PechaNotFound(f"Pecha with id {pecha_id} doesn't exist")
        repo = Repo(str(pecha_path))
        branch = _eval_branch(repo, branch)
        repo.git.checkout(branch)

    # setup auth
    storage = GithubStorage()
    setup_auth_for_old_repo(repo, org=storage.org_name, token=storage.token)

    return pecha_path


# Poti Download command
@cli.command()
@click.option("--number", "-n", help="Pecha number of pecha, for single pecha download")
@click.option(
    "--batch",
    "-b",
    help="path to text file containg list of names of \
                                     pecha in separate line. Poti batch download",
)
@click.option(
    "--filter",
    "-f",
    help="filter pecha by layer availability, specify \
                                     layer names in comma separated, eg: title,yigchung,..",
)
@click.option("--out", "-o", help="directory to store all the pecha")
def download(**kwargs):
    """
    Command to download pecha.
    If id and batch options are not provided then it will download all the pecha.
    """
    pecha_id = get_pecha_id(kwargs["number"])

    # create config dirs
    create_config_dirs()

    # get pecha
    pechas = [pecha_id]

    # download the repo
    for pecha in tqdm(pechas):
        download_pecha(pecha, kwargs["out"])

    # print location of data
    msg = f"Downloaded {pecha_id} ... ok"
    click.echo(INFO.format(msg))


# OpenPecha Formatter
formatter_types = ["ocr", "hfml(default)", "tsadra"]


@cli.command()
@click.option(
    "--name", "-n", type=click.Choice(formatter_types), help="Type of formatter"
)
@click.option("--id", "-i", type=int, help="Id of the pecha")
@click.option("--output_path", "-o", help="output path to store opf pechas")
@click.argument("input_path")
def format(**kwargs):
    """
    Command to format pecha into opf
    """
    if kwargs["name"] == "ocr":
        formatter = GoogleOCRFormatter(kwargs["output_path"])
    elif kwargs["name"] == "tsadra":
        formatter = HFMLFormatter(kwargs["output_path"])
    else:
        formatter = HFMLFormatter(kwargs["output_path"])

    formatter.create_opf(kwargs["input_path"], kwargs["id"])

export_types = ["hfml(default)", "epub"]


@cli.command()
@click.option(
    "--name", "-n", type=click.Choice(export_types), help="Type of export format"
)
@click.option("--opf_path", "-op")
@click.option("--output_path", "-o")
def export(**kwargs):
    """
    Command to export Pecha in epub
    """

    opf_path = kwargs["opf_path"]
    output_path = kwargs["output_path"]
    # if not opf_path:
    #     opf_path = f'{config["OP_PECHAS_PATH"]}/{pecha_id}/{pecha_id}.opf'

    if kwargs["name"] == "epub":
        serializer = EpubSerializer(opf_path)
    else:
        serializer = HFMLSerializer(opf_path)
    serializer.serialize(output_path)


def _get_bucket(bucket_type, bucket_name, n):
    if bucket_type == "github":
        catalog_config.GITHUB_BUCKET_CONFIG["catalog"]["end"] = n
        return GithubBucket(bucket_name, config=catalog_config.GITHUB_BUCKET_CONFIG)


def _get_filter_strategy_caller(filter_strategy):
    if filter_strategy == "non_words_ratio":
        try:
            from bonltk.text_quality import non_words_ratio
        except Exception:
            msg = (
                "bonltk not installed. Install it with `pip install bonltk` "
                "or from https://github.com/10zinten/bonltk"
            )
            raise ImportError(msg)
        return non_words_ratio


def _save_text(text, output_path, parent_path, fn):
    pecha_path = Path(output_path) / parent_path
    pecha_path.mkdir(exist_ok=True)
    vol_path = pecha_path / fn
    vol_path.write_text(text)


@cli.command()
@click.option("--output_path", "-o", type=click.Path(exists=True), required=True)
@click.option("--bucket_type", "-bt", type=click.Choice(["github"]), default="github")
@click.option("--bucket_name", "-bn", type=str, default="OpenPecha-Data")
@click.option(
    "--filter_strategy",
    "-fs",
    type=click.Choice(["non_words_ratio"]),
    default="non_words_ratio",
)
@click.option(
    "--threshold",
    "-th",
    type=float,
    default=0.8,
    help="Determines the quality of the text (1 being the highest and 0 being the lowest)",
)
@click.option("-n", type=int, default=1, help="number of pechas to download")
@click.option("--verbose", "-v", help="verbose", is_flag=True)
@click.option("--debug", "-d", help="debug", is_flag=True)
def corpus_download(
    output_path, bucket_type, bucket_name, filter_strategy, threshold, n, verbose, debug
):
    if debug:
        logging.basicConfig(level=logging.DEBUG)
    elif verbose:
        logging.basicConfig(level=logging.INFO)

    bucket = _get_bucket(bucket_type, bucket_name, n)
    filter_strategy_caller = _get_filter_strategy_caller(filter_strategy)
    for pecha_id, base in bucket.get_all_pechas_base():
        for vol_base, vol_fn in base:
            if is_text_good_quality(
                vol_base, strategy=filter_strategy_caller, threshold=threshold
            ):
                _save_text(vol_base, output_path, pecha_id, vol_fn)


@cli.command()
@click.argument("pecha_number")
@click.argument("pecha_path")
def update_layers(**kwargs):
    """
    Update all the layers when base has been updated.
    """
    pecha_id = get_pecha_id(kwargs["pecha_number"])
    src_pecha_path = download_pecha(pecha_id)

    click.echo(INFO.format(f"Updating base of {pecha_id} ..."))
    src_opf_path = src_pecha_path / f"{pecha_id}.opf"
    dst_opf_path = Path(kwargs["pecha_path"]) / f"{pecha_id}.opf"
    pecha = PechaBaseUpdate(src_opf_path, dst_opf_path)
    pecha.update()


def get_alignment_ids(path):
    path = Path(path)
    return [line for line in path.read_text().splitlines() if line]


alignment_import_types = ["tmx"]


@cli.command()
@click.argument("path")
@click.option("--verbose", "-v", help="verbose", is_flag=True)
@click.option("--debug", "-d", help="debug", is_flag=True)
@click.option(
    "--type",
    "-t",
    type=click.Choice(alignment_import_types),
    help="Type of alignment import format. Default is tmx",
)
@click.option(
    "--tm-path", "-tm", help="path to file containing OpenPecha alignment id's"
)
def import_alignment(**kwargs):
    """
    import existing alignment into OpenPecha
    """
    try:
        from openpecha.alignment.integrations.tx import OPATransifexProject
    except Exception:
        raise ImportError(
            "install with `pip install openpecha[tx]` for transifex dependencies"
        )

    if kwargs["debug"]:
        logging.basicConfig(level=logging.DEBUG)
    elif kwargs["verbose"]:
        logging.basicConfig(level=logging.INFO)
    logging.info("Creating OpenPecha alignment")
    alignment_path = alignment.create_alignment_from_tmx(tmx_path=Path(kwargs["path"]))

    added_tms_path = []
    try:
        logging.info(f"Alignment created at {alignment_path}")
        project = OPATransifexProject(org_slug="esukhia", alignment_path=alignment_path)
        project.create()
        project.import_translation()
        logging.info("Imported translation")
        logging.info("Adding TMs")
        added_tms_path = project.add_tm(
            alignment_ids=get_alignment_ids(kwargs["tm_path"])
        )
    except Exception as e:
        print(traceback.print_exc(), e)
        logging.info("Cleaning up")
        # shutil.rmtree(str(alignment_path))
    finally:
        if alignment_path.is_dir and added_tms_path:
            logging.info("Cleaning up")
            if alignment_path.is_dir():
                shutil.rmtree(str(alignment_path))
            for path in added_tms_path:
                shutil.rmtree(str(path))


@cli.command()
@click.argument("path")
@click.option("-t", "--title", help="title of the text")
@click.option(
    "--tm-path", "-tm", help="path to file containing OpenPecha alignment id's"
)
def new_translation(**kwargs):
    """
    prepare translation project in transifex
    """
    try:
        from openpecha.alignment.integrations.tx import OPATransifexProject
    except Exception:
        raise ImportError(
            "install with `pip install openpecha[tx]` for transifex dependencies"
        )

    alignment_path = alignment.create_alignment_from_source_text(
        text_path=Path(kwargs["path"])
    )
    project = OPATransifexProject(org_slug="esukhia", alignment_path=alignment_path)
    project.create()
    project.start_new_translation()
    project.add_tm(alignment_ids=get_alignment_ids(kwargs["tm_path"]))


@cli.command()
@click.argument("pecha_path")
@click.option("--save", "-s", help="save non-words counts", is_flag=True)
def qc(**kwargs):
    """
    check quality of pecha base text
    """
    pecha = OpenPechaFS(path=kwargs["pecha_path"])

    click.echo("Counting Non Words...")
    non_words_count = NonWordsCounter(empty=True)
    for base_name in pecha.components.keys():
        text = pecha.get_base(base_name)
        non_words_count += NonWordsCounter(text)

    if kwargs["save"]:
        pecha.meta.statistics["total_words"] = non_words_count.total_words
        pecha.meta.statistics["total_non_words"] = non_words_count.total_non_words
        pecha.meta.quality["non_words_ratio"] = non_words_count.non_word_ratio
        pecha.save_meta()



import re
from pathlib import Path

from setuptools import find_packages, setup


def read(fname):
    p = Path(__file__).parent / fname
    with p.open(encoding="utf-8") as f:
        return f.read()


def get_version(prop, project):
    project = Path(__file__).parent / project / "__init__.py"
    result = re.search(
        r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), project.read_text()
    )
    return result.group(1)


setup(
    name="openpecha",
    version=get_version("__version__", "openpecha"),
    author="Esukhia developers",
    author_email="esukhiadev@gmail.com",
    description="OpenPecha Toolkit allows state of the art for distributed standoff annotations on moving texts",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    license="Apache2",
    url="https://github.com/OpenPoti/openpecha-toolkit",
    packages=find_packages(),
    install_requires=[
        "botok>=0.8.8, <1.0",
        "boto3>=1.24.50, <2.0",
        "Click>=7.1.2, <8.0",
        "diff-match-patch==20181111",
        "polib==1.1.1, <2.0",
        "PyYAML>=5.0.0, <6.0",
        "pylibyaml>=0.1.0, <2.0",
        "requests>=2.22.0, <3.0",
        "antx>=0.1.10, <2.0",
        "tqdm>=4.35.0, <5.0",
        "PyGithub>=1.43.8, <2.0",
        "GitPython>=3.1.0, <4.0",
        "bs4>=0.0.1, <1.0",
        "pyewts>=0.1.3, <1.0",
        "rdflib>=5.0.0, <6.0",
        "pydantic>=1.7.3, <2.0",
        "python-docx>=0.8.11, <1.0",
    ],
    extras_require={
        "transifex": [
            "transifex-python>=1.0, <2.0",
            "python-slugify[unidecode]>=5.0, <6.0",
        ]
    },
    python_requires=">=3.6",
    entry_points={
        "console_scripts": [
            "openpecha=openpecha.cli:cli"
        ]  # command=package.module:function
    },
)



import enum
import os
import shutil
import time
from pathlib import Path

from git import Repo
from github import Github


class Storages(enum.Enum):
    GITHUB = enum.auto()


def _getenv(name, optional):
    env = os.getenv(name)
    if env:
        return env
    elif optional:
        return env
    else:
        raise RuntimeError(f"Please set {name} environment variable")


def _get_value(value, env_name, optional=False):
    return value if value else _getenv(env_name, optional)


def get_authenticated_remote_url(url, org, token):
    old_url = url.split("//")
    auth_remote_url = f"{old_url[0]}//{org}:{token}@{old_url[1]}"
    return auth_remote_url


def setup_auth_for_new_repo(repo, org, token, remote_url):
    auth_remote_url = get_authenticated_remote_url(remote_url, org, token)
    repo.create_remote("origin", auth_remote_url)
    return repo


def is_repo_authenticated(repo) -> bool:
    remote_url = repo.remote().url
    if "@" in remote_url:
        return True
    return False


def setup_auth_for_old_repo(repo, org, token):
    remote_url = repo.remote().url
    if is_repo_authenticated(repo):
        return repo
    auth_remote_url = get_authenticated_remote_url(remote_url, org, token)
    repo.remote().set_url(auth_remote_url)
    return repo


def commit_and_push(repo, message, branch="master"):
    repo.git.add("-A")
    repo.git.commit("-m", message)
    repo.git.push("origin", branch)


class Storage:
    def add_dir(self):
        raise NotImplementedError

    def remove_dir_with_name(self):
        raise NotImplementedError

    def remove_dir_with_path(self):
        raise NotImplementedError


class GithubStorage(Storage):
    """class representing Github Storage

    This storage create, update, and delete github repo and files

    Args:
        org (str, optional): github organization name
        token (str, optional): github oauth token
        username (str, optional): github username
        email (str, optional): github linked email

    """

    def __init__(
        self,
        org: str = None,
        token: str = None,
        username: str = None,
        email: str = None,
    ):
        self.org_name = _get_value(org, "OPENPECHA_DATA_GITHUB_ORG")
        self.token = _get_value(token, "GITHUB_TOKEN")
        self._username = _get_value(username, "GITHUB_USERNAME", optional=True)
        self._email = _get_value(email, "GITHUB_EMAIL", optional=True)
        self.org = Github(self.token).get_organization(self.org_name)

    @property
    def username(self):
        if not self._username:
            raise RuntimeError(
                "Please set usename attr or env variable GITHUB_USERNAME."
            )
        return self._username

    @property
    def email(self):
        if not self._email:
            raise RuntimeError("Please set email attr or env variable GITHUB_EMAIL.")
        return self._email

    def _init_local_repo(self, path: Path, remote_url: str):
        repo = Repo.init(path)
        repo = setup_auth_for_new_repo(repo, self.org_name, self.token, remote_url)
        repo.config_writer().set_value("user", "name", self.username).release()
        repo.config_writer().set_value("user", "email", self.email).release()
        return repo

    def _init_remote_repo(self, name: str, description: str):
        """Creates remote repo in Github and returns it's url."""

        repo = self.org.create_repo(name)
        time.sleep(2)
        return repo._html_url.value

    def add_dir(self, path: Path, description: str):
        """dir local dir to github."""
        remote_repo_url = self._init_remote_repo(
            name=path.name, description=description
        )
        local_repo = self._init_local_repo(path=path, remote_url=remote_repo_url)
        commit_and_push(repo=local_repo, message="Initial commit")
        return local_repo

    def remove_dir_with_name(self, name: str):
        repo = self.org.get_repo(name)
        repo.delete()

    def remove_dir_with_path(self, path: Path):
        """Remove repo with local path, assumes that local and remote name is same."""
        repo = self.org.get_repo(path.name)
        repo.delete()
        shutil.rmtree(str(path))

    def get_dir_with_name(self, name: str):
        repo = self.org.get_repo(name)
        return repo

    def get_dir_with_path(self, path: Path):
        repo = self.org.get_repo(path.name)
        return repo

    def add_file(
        self, dir_name: str, path: str, content: str, message: str, update=False
    ):
        """add file to `name` repo.

        Args:
            dir_name: name of repo to add file to.
            path: path to add the file.
            content: content of the file.
            message: git commit message.
        """

        repo = self.org.get_repo(dir_name)
        if update:
            old_content = repo.get_contents(path)
            repo.update_file(old_content.path, message, content, old_content.sha)
        else:
            repo.create_file(path, message, content)

    def get_file(self, dir_name, path: str, branch="master"):
        repo = self.org.get_repo(dir_name)
        contents = repo.get_contents(path, ref=branch)
        return contents

    def remove_file(
        self, dir_name: str, path: str, message: str, branch: str = "master"
    ):
        repo = self.org.get_repo(dir_name)
        contents = repo.get_contents(path, ref=branch)
        repo.delete_file(contents.path, message, contents.sha, branch=branch)



import re
from pathlib import Path

from ..utils import Vol2FnManager
from .formatter import BaseFormatter
from .layers import *
from .layers import AnnType, _attr_names


class PedurmaFormatter(BaseFormatter):
    """
    Pedurma formatter is to format preview of reconstructed pedurma.
    """

    def __init__(self, output_path=None, metadata=None):
        super().__init__(output_path, metadata)
        self.page = []
        self.durchen = []
        self.base_text = ""

    def text_preprocess(self, text):
        if text[0] == "\ufeff":
            text = text[1:]
        return text

    def _load_metadata(self):
        if self.metadata:
            return self.metadata
        meta_fn = self.dirs["opf_path"] / "meta.yml"
        if meta_fn.is_file():
            return self.load(meta_fn)
        else:
            return {}

    def _save_metadata(self, **kwargs):
        meta_fn = self.dirs["opf_path"] / "meta.yml"
        if kwargs:
            self.metadata.update(kwargs)
        if "id" not in self.metadata:
            self.metadata["id"] = f"opecha:{self.pecha_path.name}"
        self.dump(self.metadata, meta_fn)

    def get_input(self, input_path):
        fns = list(input_path.iterdir())
        fns_len = len(fns)
        for fn in sorted(fns):
            yield self.text_preprocess(fn.read_text()), fn.name, fns_len

    def search_before(self, ann, page_content):
        length_before = 0
        match_list = re.finditer(
            "<.*?>", page_content
        )  # list of match object of given pattern in line
        for match in match_list:
            if ann.start() > match.start():
                length_before = length_before + len(match[0])
        return length_before

    def get_pages(self, vol_text):
        result = []
        pg_text = ""
        pages = re.split(r"(〔[𰵀-󴉱]?\d+〕)", vol_text)
        for i, page in enumerate(pages[1:]):
            if i % 2 == 0:
                pg_text += page
            else:
                pg_text += page
                result.append(pg_text)
                pg_text = ""
        return result

    def parse_pagination(self, walker, base_page, page_content):
        page_ann = {}
        page_pat = re.search(r"〔[𰵀-󴉱]?(\d+)〕", page_content)
        imgnum = page_pat.group(1)
        start_page = walker + 1
        end_page = start_page + len(base_page) - 2
        span = Span(start_page, end_page)
        page_ann = Page(span, imgnum=imgnum)
        return page_ann

    def get_pub_wise_note(self, note):
        reformat_notes = {"«པེ་»": "", "«སྣར་»": "", "«སྡེ»": "", "«ཅོ་»": ""}
        note_parts = re.split("(«.+?»)", note)
        pubs = note_parts[1::2]
        notes = note_parts[2::2]
        for walker, (pub, note_part) in enumerate(zip(pubs, notes)):
            if note_part:
                reformat_notes[pub] = note_part.replace(">", "")
            else:
                if notes[walker + 1]:
                    reformat_notes[pub] = notes[walker + 1].replace(">", "")
                else:
                    reformat_notes[pub] = notes[walker + 2].replace(">", "")

        return reformat_notes

    def parse_note(self, note, walker, page_content):
        page_content = re.sub(r"〔[𰵀-󴉱]?\d+〕", "", page_content)
        note_ann = {}
        note_pat = re.search(rf"(:\S+)?{note}", page_content)
        if note_pat.group(1):
            ann_start = note_pat.start() + walker
            ann_end = ann_start + len(note_pat.group(1))
        else:
            if re.search(rf"\S+་([^#]\S+་?){note}", page_content):
                note_pat = re.search(rf"\S+་([^#]\S+་?){note}", page_content)
                grp_1_loc = page_content.find(note_pat.group(1))
            else:
                note_pat = re.search(rf"([^#]\S+་?){note}", page_content)
                grp_1_loc = note_pat.start()
            ann_start = grp_1_loc + walker
            if note_pat.group(1):
                ann_end = ann_start + len(note_pat.group(1))
            else:
                ann_end = ann_start
        note_ann = {
            "span": {
                "start": ann_start,  # the variant unit or variant location is capture with help of this span
                "end": ann_end - 1,
            },
            "variants": self.get_pub_wise_note(note),
            "collation_note": note,
        }
        page_content = re.sub(note, "", page_content, 1)
        return note_ann, page_content

    def parse_page(self, page, cur_vol_notes, cur_vol_pages, char_walker):
        page_content = page.replace("\n", "#")
        notes = re.findall(r"\<.*?\>", page_content)
        for note in notes:
            note_info, page_content = self.parse_note(note, char_walker, page_content)
            cur_vol_notes.append(("", note_info))
        new_page = self.base_extract(page)
        page_ann = self.parse_pagination(char_walker, new_page, page)
        cur_vol_pages.append(("", page_ann))
        return new_page

    def base_extract(self, text):
        text = re.sub(r"〔[𰵀-󴉱]?\d+〕", "", text)
        return re.sub("<.*?>", "", text)

    def build_layers(self, text):
        char_walker = 0
        cur_vol_notes = []
        cur_vol_pages = []
        pages = self.get_pages(text)
        for page in pages:
            new_page = self.parse_page(page, cur_vol_notes, cur_vol_pages, char_walker)
            char_walker += len(new_page)
        self.page.append(cur_vol_pages)
        self.durchen.append(cur_vol_notes)
        self.base_text = self.base_extract(text)

    def get_result(self):
        result = {
            AnnType.topic: [],
            AnnType.sub_topic: [],
            AnnType.pagination: self.page,
            AnnType.pedurma_note: self.durchen,
        }
        return result

    def get_base_text(self):
        base_text = self.base_text
        self.base_text = ""
        return base_text

    def create_opf(self, input_path, id_=None, **kwargs):
        input_path = Path(input_path)
        self._build_dirs(input_path, id_=id_)
        self.metadata = self._load_metadata()
        vol2fn_manager = Vol2FnManager(self.metadata)

        for m_text, vol_name, n_vol in self.get_input(input_path):
            vol_id = vol2fn_manager.get_vol_id(vol_name)
            print(f"[INFO] parsing Vol {vol_name} ...")
            self.build_layers(m_text)
            if "is_text" in kwargs:
                if kwargs["is_text"]:
                    continue

            base_text = self.get_base_text()
            (self.dirs["opf_path"] / "base" / f"{vol_id}.txt").write_text(base_text)

        # save pecha layers
        layers = self.get_result()
        for vol_layers, vol_id in self.format_layer(layers):
            if vol_id:
                print(f"[INFO] Creating layers for {vol_id} ...")
                vol_layer_path = self.dirs["layers_path"] / vol_id
                vol_layer_path.mkdir(exist_ok=True)
            else:
                print("[INFO] Creating index layer for Pecha ...")

            for layer, ann in vol_layers.items():
                if layer == "index":
                    layer_fn = self.dirs["opf_path"] / f"{layer}.yml"
                else:
                    layer_fn = vol_layer_path / f"{layer}.yml"
                self.dump(ann, layer_fn)

        self._save_metadata(vol2fn=dict(vol2fn_manager.vol2fn))



import re

from openpecha.core.layer import LayerEnum
from openpecha.utils import load_yaml

from .serialize import Serialize
from .epub import TsadraTemplateCSSClasses


class AnnotationTemplate:
    span_EP = "</span>"
    para_EP = "</p>"
    book_title_SP = '<p><span class="book-title"'
    sub_title_SP = '<p><span class="sub-title"'
    book_number_SP = '<p><span class="book-number"'
    author_SP = '<p><span class="author"'
    chapter_SP = '<p><span class="chapter"'
    tsawa_SP = '<span class="root-text"'
    tsawa_verse_SP = '<span class="root-text-verse"'
    citation_SP = '<span class="citation"'
    citation_verse_SP = '<span class="citation-verse"'
    citation_prose_SP = '<span class="citation-prose"'
    sabche_SP = '<span class="sabche"'
    yigchung_SP = '<span class="yigchung"'
    footnote_marker_SP = '<span class="footnote-marker"'
    footnote_EP = "</span></a>"
    footnote_reference_SP = '<span class="footnote-reference"'


class EditorSerializer(Serialize):
    def __get_adapted_span(self, span, base_id):
        """Adapts the annotation span to base-text of the text

        Adapts the annotation span, which is based on volume base-text
        to text base-text.

        Args:
            span (dict): span of a annotation, eg: {start:, end:}
            base_id (str): id of vol, where part of the text exists.

        Returns:
            adapted_start (int): adapted start based on text base-text
            adapted_end (int): adapted end based on text base-text

        """
        adapted_start = span["start"] - self.text_spans[base_id]["start"]
        adapted_end = span["end"] - self.text_spans[base_id]["start"]
        return adapted_start, adapted_end

    def get_citation_sp(self, css_class_name, ann_id):
        """Return start payload for citation according to class name

        Args:
            css_class_name (str): css class name to differentiate different types of citations

        Returns:
            str: citation start payload
        """
        start_payload = f'{AnnotationTemplate.citation_SP} id="{ann_id}">'
        if css_class_name:
            css_class_enum = TsadraTemplateCSSClasses(css_class_name)
            if css_class_enum == TsadraTemplateCSSClasses.citation_verse:
                start_payload = f'{AnnotationTemplate.citation_verse_SP} id="{ann_id}">'
            elif css_class_enum == TsadraTemplateCSSClasses.citation_prose:
                start_payload = f'{AnnotationTemplate.citation_prose_SP} id="{ann_id}">'
        return start_payload


    def get_tsawa_sp(self, css_class_name, ann_id):
        """Return start payload for tsawa according to class name

        Args:
            css_class_name (str): css class name to differentiate different types of tsawa

        Returns:
            str: tsawa start payload
        """
        start_payload = f'{AnnotationTemplate.tsawa_SP} id="{ann_id}">'
        if css_class_name:
            css_class_enum = TsadraTemplateCSSClasses(css_class_name)
            if css_class_enum == TsadraTemplateCSSClasses.tsawa_verse:
                start_payload = f'{AnnotationTemplate.tsawa_verse_SP} id="{ann_id}">'
        return start_payload
        
    def apply_annotation(self, base_id, ann, uuid2localid):
        """Applies annotation to specific volume base-text, where part of the text exists.

        Args:
            base_id (str): id of vol, where part of the text exists.
            ann (dict): annotation of any type.

        Returns:
            None

        """
        only_start_ann = False
        start_payload = ""
        end_payload = ""
        ann_id = ann["id"]
        if ann["type"] == LayerEnum.correction.value:
            start_payload = "("
            end_payload = f',{ann["correction"]})'
        elif ann["type"] == LayerEnum.peydurma.value:
            start_payload = "#"
            only_start_ann = True
        elif ann["type"] == LayerEnum.error_candidate.value:
            start_payload = "["
            end_payload = "]"
        elif ann["type"] == LayerEnum.book_title.value:
            start_payload = f'{AnnotationTemplate.book_title_SP} id="{ann_id}">'
            end_payload = AnnotationTemplate.span_EP + AnnotationTemplate.para_EP
        elif ann["type"] == LayerEnum.sub_title.value:
            start_payload = f'{AnnotationTemplate.sub_title_SP} id="{ann_id}">'
            end_payload = AnnotationTemplate.span_EP + AnnotationTemplate.para_EP
        elif ann["type"] == LayerEnum.book_number.value:
            start_payload = f'{AnnotationTemplate.book_number_SP} id="{ann_id}">'
            end_payload = AnnotationTemplate.span_EP + AnnotationTemplate.para_EP
        elif ann["type"] == LayerEnum.author.value:
            start_payload = f'{AnnotationTemplate.author_SP} id="{ann_id}">'
            end_payload = AnnotationTemplate.span_EP + AnnotationTemplate.para_EP
        elif ann["type"] == LayerEnum.chapter.value:
            start_payload = f'{AnnotationTemplate.chapter_SP} id="{ann_id}">'
            end_payload = AnnotationTemplate.span_EP + AnnotationTemplate.para_EP
        elif ann["type"] == LayerEnum.tsawa.value:
            css_class_name = self.get_css_class_name(ann)
            start_payload = self.get_tsawa_sp(css_class_name, ann_id)
            end_payload = AnnotationTemplate.span_EP
        elif ann["type"] == LayerEnum.citation.value:
            css_class_name = self.get_css_class_name(ann)
            start_payload = self.get_citation_sp(css_class_name, ann_id)
            end_payload = AnnotationTemplate.span_EP
        elif ann["type"] == LayerEnum.sabche.value:
            start_payload = f'{AnnotationTemplate.sabche_SP} id="{ann_id}">'
            end_payload = AnnotationTemplate.span_EP
        elif ann["type"] == LayerEnum.yigchung.value:
            start_payload = f'{AnnotationTemplate.yigchung_SP} id="{ann_id}">'
            end_payload = AnnotationTemplate.span_EP
        elif ann["type"] == LayerEnum.footnote.value:
            start_payload = f'<a href="#fr{ann_id}>{AnnotationTemplate.footnote_marker_SP} id="fm{ann_id}">'
            end_payload = AnnotationTemplate.footnote_EP

        start_cc, end_cc = self.__get_adapted_span(ann["span"], base_id)
        self.add_chars(base_id, start_cc, True, start_payload)
        if not only_start_ann:
            self.add_chars(base_id, end_cc, False, end_payload)

    def get_footnote_references(self, footnotes):
        """Generate footnote references using footnote layer

        Args:
            footnotes (dict): footnote annotation information

        Returns:
            str: footnote references p tags
        """
        footnote_references = ""
        for footnote_id, footnote in footnotes.items():
            footnote_references += f'<p><a href="#fm{footnote_id}">{AnnotationTemplate.footnote_reference_SP} id="fr{footnote_id}">{footnote["footnote_ref"]}</span></a></p>'
        return footnote_references

    def p_tag_adder(self, body_text):
        """Add p tag to lines where it is missing

        Args:
            body_text (str): body tag of editor after layers are applied

        Returns:
            str: body tag with proper p tags
        """
        new_body_text = ""
        body_text = re.sub(r"\n</span>", "\n</span>\n", body_text)
        paras = body_text.split("\n")
        para_flag = False
        cur_para = ""
        id_walker = 1
        cur_span_payload = ""
        for para in paras:
            if "<p" not in para:
                if re.search("<span.+?</span>", para):
                    new_body_text += f"<p>{para}</p>"
                elif "<span" in para and not para_flag:
                    cur_span_payload = re.search("<span .+>", para)[0]
                    cur_para += f"<p>{para}</span></p>"
                    para_flag = True
                elif para_flag and "</span>" not in para:
                    cur_para += (
                        f'<p>{cur_span_payload[:-2]}{id_walker}">{para}</span></p>'
                    )
                    id_walker += 1
                elif "</span>" in para:
                    cur_para += f'<p>{cur_span_payload[:-2]}{id_walker}">{para}</p>'
                    new_body_text += cur_para
                    cur_para = ""
                    para_flag = False
                    cur_span_payload = ""
                    id_walker = 1
                else:
                    new_body_text += f"<p>{para}</p>"
            else:
                new_body_text += para
        return new_body_text

    def serialize(self):
        """Opf is serialize to html format in order to present it in editor workspace

        Yields:
            str, str: base file name, serialized html of that base file
        """
        self.apply_layers()
        self.layers = [layer for layer in self.layers if layer != "Pagination"]

        results = self.get_result()
        for base_name, result in results.items():
            footnote_ref_tag = ""
            if "Footnote" in self.layers:
                footnote_fn = self.opf_path / "layers" / base_name / "Footnote.yml"
                footnote_layer = load_yaml(footnote_fn)
                footnote_ref_tag = self.get_footnote_references(
                    footnote_layer["annotations"]
                )
            result = self.p_tag_adder(result)
            result = f"<html>\n<head>\n<title></title>\n</head>\n<body>\n{result}{footnote_ref_tag}</body>\n</html>"
            yield base_name, result



import re
from pathlib import Path

from bs4 import BeautifulSoup


def parse_segment(multilingual_segment, src_lang, tar_lang):
    src_seg = ""
    tar_seg = ""
    segments = multilingual_segment.find_all("tuv")
    for segment in segments:
        seg_text = segment.find("seg").text if segment.find("seg").text else "-----"
        seg_text = re.sub(r"\n|\t", " ", seg_text)
        seg_text = re.sub(r"\s{2,}", " ", seg_text)
        seg_text = seg_text.strip() + "\n"
        if segment.get("xml:lang", "") == src_lang:
            src_seg = seg_text
        elif segment.get("xml:lang", "") == tar_lang:
            tar_seg = seg_text
    return [src_seg, tar_seg]


def parse_tmx_text(tmx_content, src_lang="bo", tar_lang="en"):
    tmx_text = {src_lang: "", tar_lang: ""}
    soup = BeautifulSoup(tmx_content, "xml")
    multilingual_segments = soup.find_all("tu")
    for multilingual_segment in multilingual_segments:
        src_seg, tar_seg = parse_segment(multilingual_segment, src_lang, tar_lang)
        tmx_text[src_lang] += src_seg
        tmx_text[tar_lang] += tar_seg
    return tmx_text


def normalise_dict_keys(dict_obj):
    """replaces `:` to `-` in dict"""

    result = {}
    for key, value in dict_obj.items():
        key = key.replace(":", "-")
        result[key] = value
    return result


def parse_metadata(tmx):
    soup = BeautifulSoup(tmx, "xml")
    source_metadata = soup.header.attrs
    source_metadata = normalise_dict_keys(source_metadata)
    return source_metadata


def parse_tmx(tmx_path):
    tmx = tmx_path.read_text(encoding="utf-8-sig")
    source_metadata = parse_metadata(tmx)
    src_lang = source_metadata.get("srclang", "")
    tar_lang = source_metadata.get("adminlang", "")
    if re.search(r"en", tar_lang):
        tar_lang = "en"
    tmx_text = parse_tmx_text(tmx, src_lang, tar_lang)
    src_text = tmx_text[src_lang]
    tar_text = tmx_text[tar_lang]
    tar_text = re.sub(r"\d+", "", tar_text)
    tar_text = tar_text.replace("[]", "")
    tar_text = tar_text.replace("()", "")
    tar_text = tar_text.replace(r"{}", "")
    return src_text, tar_text, source_metadata
